#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from matplotlib.ticker import MultipleLocator, FormatStrFormatter,ScalarFormatter
import matplotlib.colors as colors
import numpy as np
from copy import deepcopy
from ..ANTARESS_general.utils import stop


#%% Symbols
degree_sign= u'\N{DEGREE SIGN}'


#%% Routines

def custom_axis(plt,ax=None,fig = None, x_range=None,y_range=None,z_range=None,position=None,colback=None,
		    x_mode=None,y_mode=None,z_mode=None,
		    x_title=None,y_title=None,z_title=None,x_title_dist=None,y_title_dist=None,z_title_dist=None,
		    no_xticks=False,no_yticks=False,no_zticks=True,
		    top_xticks=None,right_yticks=None,   
            hide_xticks = False,hide_yticks = False,		
		    xmajor_int=None,xminor_int=None,ymajor_int=None,yminor_int=None,zmajor_int=None,zminor_int=None,
		    xmajor_form=None,ymajor_form=None,zmajor_form=None,
		    xmajor_length=None,ymajor_length=None,zmajor_length=None,xminor_length=None,yminor_length=None,zminor_length=None,
		    font_size=None,font_thick=None,xfont_size=None,yfont_size=None,zfont_size=None,
		    axis_thick=None,xmajor_thick=None,xminor_thick=None,ymajor_thick=None,yminor_thick=None,zmajor_thick=None,zminor_thick=None,
		    dir_x=None,dir_y=None,dir_z=None,
            xtick_pad=None,ytick_pad=None,ztick_pad=None,
            xlab_col=None,ylab_col=None,zlab_col=None,
            hide_axis=None,
            right_axis=False,secy_title=None,secy_range=None,secy_title_dist=None,no_secyticks=None,secymajor_int=None,dir_secy=None,secyfont_size=None,
            secymajor_length=None,secymajor_thick=None,secyminor_length=None,secyminor_thick=None,secymajor_form=None,secyminor_int=None,secylab_col=None):
    r"""**Plot axis.**
    
    General routines to set up plot axis to default or user-selected values.  
    
    Args:
        TBD
    
    Returns:
        None
    
    """ 
		
    #Font
    #    - thick: 'light', 'normal', 'medium', 'semibold', 'bold', 'heavy', 'black'					
    font_size_loc=font_size if font_size is not None else 10.
    font_thick_loc=font_thick if font_thick is not None else 'normal'
    plt.rc('font', size=font_size_loc,weight=font_thick_loc,**{'family':'sans-serif','sans-serif':['Helvetica']})
# 	plt.rc('text', usetex=True)
    plt.rcParams['pdf.fonttype'] = 42

    #Axis
    if ax==None:ax=plt.gca()

    #Axis frame position
    #    - corresponds to the corners of the image
    if position is not None:
        if fig is None:plt.subplots_adjust(left=position[0],bottom=position[1],right=position[2],top=position[3]) 
        else:fig.subplots_adjust(left=position[0],bottom=position[1],right=position[2],top=position[3]) 

    #Axis ranges
    if x_range is not None:ax.set_xlim([x_range[0], x_range[1]])
    else:x_range = ax.get_xlim()
    dx_range = x_range[1]-x_range[0]
    if y_range is not None:ax.set_ylim([y_range[0], y_range[1]])
    else:y_range = ax.get_ylim()
    dy_range = y_range[1]-y_range[0]
    if z_range is not None:ax.set_zlim([z_range[0], z_range[1]])

    #Set axis to log mode if required
    if x_mode=='log':
        ax.set_xscale('log')
    if y_mode=='log':
        ax.set_yscale('log')
    if z_mode=='log':
        ax.set_zscale('log')


    #Axis titles	
    xfont_size_loc=xfont_size if xfont_size is not None else 10.
    yfont_size_loc=yfont_size if yfont_size is not None else 10. 
    zfont_size_loc=zfont_size if zfont_size is not None else 10. 
    if x_title is not None:
            ax.set_xlabel(x_title,fontsize=xfont_size_loc,weight=font_thick_loc)
    if y_title is not None:
            if (right_yticks=='on'):  #set title to right axis          
                ax.set_ylabel(y_title,fontsize=yfont_size_loc,rotation=270,labelpad=22,weight=font_thick_loc)
            else:            
                ax.set_ylabel(y_title,fontsize=yfont_size_loc,weight=font_thick_loc) 
    if z_title is not None:   
            ax.set_zlabel(z_title,fontsize=zfont_size_loc,weight=font_thick_loc)
                
    #Axis title distance
    if x_title_dist is not None:ax.xaxis.labelpad = x_title_dist
    if y_title_dist is not None:ax.yaxis.labelpad = y_title_dist
    if z_title_dist is not None:ax.zaxis.labelpad = z_title_dist
    	
    #Axis background color
    if colback is not None:ax.set_facecolor(colback)

    #Axis thickness
    axis_thick_loc=axis_thick if axis_thick is not None else 1.
    if ('bottom' in list(ax.spines.keys())):ax.spines['bottom'].set_linewidth(axis_thick_loc)
    if ('top' in list(ax.spines.keys())):   ax.spines['top'].set_linewidth(axis_thick_loc)
    if ('left' in list(ax.spines.keys())):  ax.spines['left'].set_linewidth(axis_thick_loc)
    if ('right' in list(ax.spines.keys())): ax.spines['right'].set_linewidth(axis_thick_loc)

    #X ticks and label on top
    if (top_xticks=='on'):
        ax.xaxis.tick_top()
        ax.xaxis.set_label_position('top') 	
    #Y ticks and label on right
    if (right_yticks=='on'):
         ax.yaxis.tick_right()
         ax.yaxis.set_label_position('right')
  
    #------------------------------------------------------------------------
    #X ticks (on by default)		
    if (no_xticks==False):
      
        #Interval between major ticks	
        if xmajor_int is not None:
            n_ticks = int(dx_range/xmajor_int)
            if n_ticks>100:xmajor_int = dx_range/100.
            ax.xaxis.set_major_locator(MultipleLocator(xmajor_int))

		#Direction of ticks
        if dir_x==None:dir_x='in' 

	    #Major ticks
        xmajor_length_loc=xmajor_length if xmajor_length is not None else 7
        xmajor_thick_loc=xmajor_thick if xmajor_thick is not None else 1.5		
        xtick_pad_loc=xtick_pad if xtick_pad is not None else 5  
        ax.tick_params('x', length=xmajor_length_loc, which='major',width=xmajor_thick_loc,
					  direction=dir_x, pad=xtick_pad_loc,labelsize=xfont_size_loc,top=True)
       
        #Minor ticks		
        xminor_length_loc=xminor_length if xminor_length is not None else xmajor_length_loc/2.	
        xminor_thick_loc=xminor_thick if xminor_thick is not None else 1.5		
        ax.tick_params('x', length=xminor_length_loc, which='minor',width=xminor_thick_loc,
					  direction=dir_x,labelsize=xfont_size_loc,top=True)
        
        #Major ticks label format		
        if xmajor_form is not None:ax.xaxis.set_major_formatter(FormatStrFormatter(xmajor_form))
        if x_mode=='log':
            if xmajor_form is not None:ax.xaxis.set_major_formatter(ScalarFormatter(xmajor_form))
            else:ax.xaxis.set_major_formatter(ScalarFormatter())
	
	    #Interval between minor ticks	
        if x_mode=='log':xminor_int=None
        if xminor_int is not None:ax.xaxis.set_minor_locator(MultipleLocator(xminor_int))		

        #Ticks labels color
        if xlab_col is not None:[i_col.set_color(xlab_col) for i_col in ax.get_xticklabels()]

        #Hide top ticks
        if hide_xticks:ax.xaxis.set_ticks_position('bottom')

    else:
        ax.set_xticks([])	
		
	#-----------------	
    #Y ticks (on by default)	
    if (no_yticks==False):

        #Interval between major ticks		
        if ymajor_int is not None:
            n_ticks = int(dy_range/ymajor_int)
            if n_ticks>100:ymajor_int = dy_range/100.
            ax.yaxis.set_major_locator(MultipleLocator(ymajor_int))	
       
		#Direction of ticks
        if dir_y==None:dir_y='in'         

	    #Major ticks
        ymajor_length_loc=ymajor_length if ymajor_length is not None else 7
        ymajor_thick_loc=ymajor_thick if ymajor_thick is not None else 1.5		
        ytick_pad_loc=ytick_pad if ytick_pad is not None else 5 		
        ax.tick_params('y', length=ymajor_length_loc, which='major',width=ymajor_thick_loc,
        					  direction=dir_y, pad=ytick_pad_loc,labelsize=yfont_size_loc, right=True)

	    #Minor ticks	
        yminor_length_loc=yminor_length if yminor_length is not None else ymajor_length_loc/2.
        yminor_thick_loc=yminor_thick if yminor_thick is not None else 1.5			
        ax.tick_params('y', length=yminor_length_loc, which='minor',width=yminor_thick_loc,
        					  direction=dir_y,labelsize=yfont_size_loc, right=True)

   	    #Major ticks label format		
        if ymajor_form is not None:ax.yaxis.set_major_formatter(FormatStrFormatter(ymajor_form))	
        if y_mode=='log':
            if ymajor_form is not None:ax.yaxis.set_major_formatter(ScalarFormatter(ymajor_form))
            else:ax.yaxis.set_major_formatter(ScalarFormatter())    

	    #Interval between minor ticks	
        if y_mode=='log':yminor_int=None
        if yminor_int is not None:
            n_ticks = int(ymajor_int/yminor_int)
            if n_ticks>50:yminor_int = ymajor_int/50.
            ax.yaxis.set_minor_locator(MultipleLocator(yminor_int))		

	    #Ticks labels color
        if ylab_col is not None:[i_col.set_color(ylab_col) for i_col in ax.get_yticklabels()]
        
        #Hide right ticks
        if hide_yticks:ax.yaxis.set_ticks_position('left')

    else:
        ax.set_yticks([])	

	#-----------------	
    #Z ticks (off by default)	
    if (no_zticks==False):
		
        #Interval between major ticks		
        if zmajor_int is not None:ax.zaxis.set_major_locator(MultipleLocator(zmajor_int))	

		#Direction of ticks
        if dir_z==None:dir_z='in' 

	      #Major ticks length	
        zmajor_length_loc=zmajor_length if zmajor_length is not None else 7
        zmajor_thick_loc=zmajor_thick if zmajor_thick is not None else 1.		
        ztick_pad_loc=ztick_pad if ztick_pad is not None else 5 		
        ax.tick_params('z', length=zmajor_length_loc, which='major',width=zmajor_thick_loc,
					  direction=dir_z, pad=ztick_pad_loc,labelsize=zfont_size_loc)
	
	      #Minor ticks length	
        zminor_length_loc=zminor_length if zminor_length is not None else zmajor_length_loc/2.
        zminor_thick_loc=zminor_thick if zminor_thick is not None else 1.5			
        ax.tick_params('z', length=zminor_length_loc, which='minor',width=zminor_thick_loc,
					  direction=dir_z,labelsize=zfont_size_loc)
	
	      #Major ticks label format		
        if zmajor_form is not None:
            majorFormatter = FormatStrFormatter(zmajor_form)
            ax.zaxis.set_major_formatter(majorFormatter)		
			
	      #Interval between minor ticks			
        if zminor_int is not None:	
            minorLocator   = MultipleLocator(zminor_int)
            ax.zaxis.set_minor_locator(minorLocator)

	      #Ticks labels color
        if zlab_col is not None:[i_col.set_color(zlab_col) for i_col in ax.get_zticklabels()]

    # else:
    #     ax.set_zticks([])	
		
    #------------------------------------------------------------------------
    #Secondary axis (right side)
    if right_axis==True:
         if right_yticks=='on':stop('Nominal axis already set to right axis')
         newaxvert = ax.twinx()
         newaxvert.yaxis.tick_right()
         newaxvert.yaxis.set_label_position('right') 

         #Title 
         if secy_title is not None:
             yfont_size_loc=secyfont_size if secyfont_size is not None else 10. 
             newaxvert.set_ylabel(secy_title,fontsize=yfont_size_loc,rotation=0) 
         if secy_title_dist is not None:newaxvert.yaxis.labelpad = secy_title_dist

         #Range              
         if secy_range is not None:newaxvert.set_ylim([secy_range[0], secy_range[1]])
    
    	   #-----------------	
         if (no_secyticks==None) or (no_secyticks==''):
             
             #Interval between major ticks
             if secymajor_int is not None:
                newaxvert.yaxis.set_major_locator(MultipleLocator(secymajor_int))	
                
             #Direction of ticks
             if dir_secy==None:dir_secy='in' 
    
    	      #Major ticks length	
             secymajor_length_loc=secymajor_length if secymajor_length is not None else 10
             secymajor_thick_loc=secymajor_thick if secymajor_thick is not None else 1.5			
             newaxvert.tick_params('y', length=secymajor_length_loc, which='major',width=secymajor_thick_loc,
    					  direction=dir_secy)
    	
    	      #Minor ticks length	
             secyminor_length_loc=secyminor_length if secyminor_length is not None else secymajor_length_loc/2.
             secyminor_thick_loc=secyminor_thick if secyminor_thick is not None else 1.5			
             newaxvert.tick_params('y', length=secyminor_length_loc, which='minor',width=secyminor_thick_loc,
    					  direction=dir_secy)
    	
    	      #Major ticks label format		
             if secymajor_form is not None:newaxvert.yaxis.set_major_formatter(FormatStrFormatter(secymajor_form))		
    			
    	      #Interval between minor ticks			
             if secyminor_int is not None:newaxvert.yaxis.set_minor_locator(MultipleLocator(secyminor_int))
    
    	      #Ticks labels color
             if secylab_col is not None:[i_col.set_color(secylab_col) for i_col in newaxvert.get_yticklabels()]
    
         else:
             newaxvert.set_yticks([])	

	#-------------------------------------------
      #Hide all axis
    if hide_axis:
          ax.xaxis.set_visible(False)
          ax.yaxis.set_visible(False)
          ax.spines['bottom'].set_color('white')  
          ax.spines['top'].set_color('white')  
          ax.spines['left'].set_color('white')  
          ax.spines['right'].set_color('white')  
    
              
    return None 
	

def scaled_title(sc_fact10,y_title):
    r"""**Title scaling.**
    
    Applies power-of-ten scaling to title value.  
    
    Args:
        sc_fact10 (int): power of ten scaling
        y_title (str): title
    
    Returns:
        y_title (str): title preceded by scaling value
    
    """ 
    if sc_fact10!=0.:
        sc_sign='-' if sc_fact10<0. else ''
        y_title='10$^{'+sc_sign+'%i' % abs(sc_fact10)+'}$ '+y_title
    return y_title                        



def autom_range_ext(ax_range_in,ax_min,ax_max,ext_fact=0.05):  
    r"""**Automatic range.**
    
    Defines axis range automatically.  
    
    Args:
        ax_range_in (list): axis range to be set, if user-defined
        ax_min (float): minimum value on the axis
        ax_max (float): maximum value on the axis
        ext_fact (float): extension of axis on both sides compared to min and max values
    
    Returns:
        ax_range (list): axis range
    
    """ 

    if ax_range_in is None:
        ax_range = np.array([ax_min,ax_max])   
        dx_range=ax_range[1]-ax_range[0]
        ax_range[0]-=ext_fact*dx_range
        ax_range[1]+=ext_fact*dx_range
        dx_range=ax_range[1]-ax_range[0]
    else:ax_range=ax_range_in
    return ax_range


def autom_tick_prop(dax_range):
    r"""**Automatic ticks.**
    
    Defines tick spacings and format automatically.  
    
    Args:
        dax_range (float): axis extension
    
    Returns:
        axmajor_int (float): major ticks spacing
        axminor_int (float): minor ticks spacing
        axmajor_form (str): major tick format
    
    """ 
    if   dax_range>1e11+0.1:axmajor_int,axminor_int,axmajor_form=5e10,1e10,'%.1e' 
    elif dax_range>1e10+0.1:axmajor_int,axminor_int,axmajor_form=5e9,1e9,'%.1e' 
    elif dax_range>1e9+0.1: axmajor_int,axminor_int,axmajor_form=5e8,1e8,'%.1e'     
    elif dax_range>1e8+0.1: axmajor_int,axminor_int,axmajor_form=5e7,1e7,'%.1e'     
    elif dax_range>1e7+0.1: axmajor_int,axminor_int,axmajor_form=5e6,1e6,'%.1e' 
    elif dax_range>1e6+0.1: axmajor_int,axminor_int,axmajor_form=5e5,1e5,'%.1e' 
    elif dax_range>1e5+0.1: axmajor_int,axminor_int,axmajor_form=5e4,1e4,'%.1e' 
    elif dax_range>1e4+0.1: axmajor_int,axminor_int,axmajor_form=5000.,1000.,'%.1e' 
    elif dax_range>5e3+0.1: axmajor_int,axminor_int,axmajor_form=1000.,500.,'%i'     
    elif dax_range>1e3+0.1: axmajor_int,axminor_int,axmajor_form=500.,100.,'%i' 
    elif dax_range>500.1:   axmajor_int,axminor_int,axmajor_form=200.,50.,'%i' 
    elif dax_range>200.1:   axmajor_int,axminor_int,axmajor_form=100.,10.,'%i'     
    elif dax_range>100.1:   axmajor_int,axminor_int,axmajor_form=50.,10.,'%i' 
    elif dax_range>50.1:    axmajor_int,axminor_int,axmajor_form=10.,1.,'%i' 
    elif dax_range>20.1:    axmajor_int,axminor_int,axmajor_form=5.,1.,'%i' 
    elif dax_range>15.1:    axmajor_int,axminor_int,axmajor_form=4.,1,'%i'    
    elif dax_range>10.1:    axmajor_int,axminor_int,axmajor_form=2.,1.,'%i'
    elif dax_range>6.1:     axmajor_int,axminor_int,axmajor_form=2.,1.,'%i' 
    elif dax_range>3.5:     axmajor_int,axminor_int,axmajor_form=1.,0.5,'%i'  
    elif dax_range>3.1:     axmajor_int,axminor_int,axmajor_form=1.,0.5,'%i'     
    elif dax_range>1.1:     axmajor_int,axminor_int,axmajor_form=5e-1,1e-1,'%.1f'     
    elif dax_range>7.1e-1:  axmajor_int,axminor_int,axmajor_form=2e-1,5e-2,'%.1f'     
    elif dax_range>5.1e-1:  axmajor_int,axminor_int,axmajor_form=2e-1,5e-2,'%.1f'    
    elif dax_range>3.1e-1:  axmajor_int,axminor_int,axmajor_form=1e-1,5e-2,'%.1f'     
    elif dax_range>1.1e-1:  axmajor_int,axminor_int,axmajor_form=5e-2,1e-2,'%.2f'
    elif dax_range>5.1e-2:  axmajor_int,axminor_int,axmajor_form=2e-2,1e-2,'%.2f'       
    elif dax_range>4.1e-2:  axmajor_int,axminor_int,axmajor_form=1e-2,5e-3,'%.2f'
    elif dax_range>3.1e-2:  axmajor_int,axminor_int,axmajor_form=1e-2,5e-3,'%.2f' 
    elif dax_range>1.1e-2:  axmajor_int,axminor_int,axmajor_form=5e-3,1e-3,'%.3f'
    elif dax_range>5.1e-3:  axmajor_int,axminor_int,axmajor_form=2e-3,1e-3,'%.3f'    
    elif dax_range>2.1e-3:  axmajor_int,axminor_int,axmajor_form=1e-3,5e-4,'%.3f'   
    elif dax_range>1.1e-3:  axmajor_int,axminor_int,axmajor_form=5e-4,1e-4,'%.4f'     
    elif dax_range>5.1e-4:  axmajor_int,axminor_int,axmajor_form=5e-4,1e-4,'%.4f' 
    elif dax_range>1.1e-4:  axmajor_int,axminor_int,axmajor_form=2e-4,1e-4,'%.4f' 
    else:axmajor_int,axminor_int,axmajor_form=None,None,None 
    return axmajor_int,axminor_int,axmajor_form    
    



def adjust_isosize(real_bounds,xpos0,ypos0,xpos1,ypos1,max_window_size):
    r"""**2D isotropic plot scaling.**
    
    Adjusts the size and position of the plot window, maintaining isotropy.
    
    Variables along the X and Y axis must have the same units.
    
    The lower left corner of the plot at `(xpos0,ypos0)` is taken as reference. 
    The `(x_pos1,y_pos1)` defines the maximum extension of the plot at the top right corner. 
    Depending on the respective width/height of the original plot, the larger side (within the axes) is scaled to `max_window_size` while the smallest side is then scaled by isotropy.    
    
    Args:
        TBD
    
    Returns:
        TBD
    
    """     

    #We use a square plot
    height=max_window_size					
    width=max_window_size	

    #Real ranges of the plot
    #    - real_bounds must be [min_x,max_x,min_y,max_y]
    #    - xpos0,ypos0,xpos1,ypos1 define the position of the axis limits within the plot window defined by width,height
    sub_width_set=(real_bounds[1]-real_bounds[0])
    sub_height_set=(real_bounds[3]-real_bounds[2])	
   
    #-----------------------
 
    #We calculate the position of the upper side of the axis, assuming isotropy
    #    - here we calculate the position of the axis (not the overall box) in fraction of 1
    #    - (sub_width_set) <-> (xpos1-xpos0) and (sub_height_set) <-> (ypos1-ypos0)		
    ypos1_temp=ypos0+(xpos1-xpos0)*(sub_height_set/sub_width_set)	
 
    #If the new upper side is higher than the required limit
    if (ypos1_temp>ypos1):   
	    #We set the upper limit to ypos1 and adjust instead the right side  
	    #    - in fraction of 1				
	    xpos1=xpos0+(ypos1-ypos0)*(sub_width_set/sub_height_set)			
    else:
	    ypos1=ypos1_temp
					
    return width,height,[xpos0,ypos0,xpos1,ypos1]


def adjust_3D_isosize(ax,x_range,y_range,z_range):
    r"""**3D isotropic plot scaling.**
    
    Adjusts the size of the plot window, maintaining isotropy.
      
    
    Args:
        TBD
    
    Returns:
        TBD

    """  
    max_range = np.array([x_range[1]-x_range[0],y_range[1]-y_range[0], z_range[1]-z_range[0]]).max() / 2.0	
    mean_x = 0.5*(x_range[1]+x_range[0])
    mean_y = 0.5*(y_range[1]+y_range[0])
    mean_z = 0.5*(z_range[1]+z_range[0])
    ax.set_xlim(mean_x - max_range, mean_x + max_range)
    ax.set_ylim(mean_y - max_range, mean_y + max_range)
    ax.set_zlim(mean_z - max_range, mean_z + max_range)
    
    return ax							


def truncate_colormap(cmap, minval=0.0, maxval=1.0, n=100):
    r"""**Colormap truncation.**
    
    Limits colormap to the chosen range for the plotted variable.
    
    Args:
        TBD
    
    Returns:
        TBD
    
    """   
    new_cmap = colors.LinearSegmentedColormap.from_list('trunc({n},{a:.2f},{b:.2f})'.format(n=cmap.name, a=minval, b=maxval),
        cmap(np.linspace(minval, maxval, n)))
    return new_cmap


def stackrel(val,sub,sup,form):
    r"""**Stackrel truncation.**
    
    Print value with upper and lower error bars.
    
    Args:
        TBD
    
    Returns:
        TBD
    
    """ 
    form = deepcopy("{"+form+"}")
    return r"$"+form.format(val)+"\genfrac{}{}{0}{}{+"+form.format(sup)+"}{-"+form.format(sub)+"}$"



def mscatter(plt ,x, y, ax=None, m=None, **kw):
    r"""**Multiple markers.**
    
    Overrides plt.scatter to have various markers in one command.
    
    Args:
        TBD
    
    Returns:
        TBD
    
    """ 
    import matplotlib.markers as mmarkers

    if not ax: ax = plt.gca()

    sc = ax.scatter(x, y, **kw)
    
    if (m is not None) and (len(m) == len(x)):
        paths = []
        for marker in m:
            if isinstance(marker, mmarkers.MarkerStyle):
                marker_obj = marker
            else:
                marker_obj = mmarkers.MarkerStyle(marker)
            path = marker_obj.get_path().transformed(marker_obj.get_transform())
            paths.append(path)
        sc.set_paths(paths)
        
    return sc





def plot_shade_range(ax,shade_range,x_range_loc,y_range_loc,mode='fill',facecolor='grey',zorder=-1,alpha=0.2,compl=False):
    r"""**Shaded ranges.**
    
    Shades list of ranges.
    
    Args:
        TBD
    
    Returns:
        TBD
    
    """ 
    for i_int,bd_int in enumerate(shade_range):
        if compl:
            if (i_int==0 and bd_int[0]>x_range_loc[0]):
                bd_int_loc=[x_range_loc[0],bd_int[0]] #shade area before first interval                 
                if mode=='span':ax.axvspan(bd_int_loc[0],bd_int_loc[1], facecolor=facecolor, alpha=alpha,zorder=zorder)
                elif mode=='fill':ax.fill([bd_int_loc[0],bd_int_loc[1],bd_int_loc[1],bd_int_loc[0]],[y_range_loc[0],y_range_loc[0],y_range_loc[1],y_range_loc[1]], fill=True,color='dodgerblue',alpha=alpha,zorder=zorder,ls='')  
                elif mode=='hatch':ax.fill([bd_int_loc[0],bd_int_loc[1],bd_int_loc[1],bd_int_loc[0]],[y_range_loc[0],y_range_loc[0],y_range_loc[1],y_range_loc[1]], fill=False, hatch='\\',color=facecolor,zorder=zorder)         
            if (i_int==len(shade_range)-1 and bd_int[1]<x_range_loc[1]):
                bd_int_loc=[bd_int[1],x_range_loc[1]] #shade area after last interval
                if mode=='span':ax.axvspan(bd_int_loc[0],bd_int_loc[1], facecolor=facecolor, alpha=alpha,zorder=zorder)
                elif mode=='fill':ax.fill([bd_int_loc[0],bd_int_loc[1],bd_int_loc[1],bd_int_loc[0]],[y_range_loc[0],y_range_loc[0],y_range_loc[1],y_range_loc[1]], fill=True,color='dodgerblue',alpha=alpha,zorder=zorder,ls='')
                elif mode=='hatch':ax.fill([bd_int_loc[0],bd_int_loc[1],bd_int_loc[1],bd_int_loc[0]],[y_range_loc[0],y_range_loc[0],y_range_loc[1],y_range_loc[1]], fill=False, hatch='\\',color=facecolor,zorder=zorder)      
            if i_int>0:
                bd_int_loc=[shade_range[i_int-1][1],bd_int[0]]          #shade area between current and previous interval
                if mode=='span':ax.axvspan(bd_int_loc[0],bd_int_loc[1], facecolor=facecolor, alpha=alpha,zorder=zorder)
                elif mode=='fill':ax.fill([bd_int_loc[0],bd_int_loc[1],bd_int_loc[1],bd_int_loc[0]],[y_range_loc[0],y_range_loc[0],y_range_loc[1],y_range_loc[1]], fill=True,color='dodgerblue',alpha=alpha,zorder=zorder,ls='')
                elif mode=='hatch':ax.fill([bd_int_loc[0],bd_int_loc[1],bd_int_loc[1],bd_int_loc[0]],[y_range_loc[0],y_range_loc[0],y_range_loc[1],y_range_loc[1]], fill=False, hatch='\\',color=facecolor,zorder=zorder)        
        else:
            if (bd_int[1]>x_range_loc[0]) & (bd_int[0]<x_range_loc[1]):
                bd_int_loc=[np.max([x_range_loc[0],bd_int[0]]),np.min([x_range_loc[1],bd_int[1]])]
                if mode=='span':ax.axvspan(bd_int_loc[0],bd_int_loc[1], facecolor=facecolor, alpha=alpha,zorder=zorder)
                elif mode=='fill':ax.fill([bd_int_loc[0],bd_int_loc[1],bd_int_loc[1],bd_int_loc[0]],[y_range_loc[0],y_range_loc[0],y_range_loc[1],y_range_loc[1]], fill=True,color='dodgerblue',alpha=alpha,zorder=zorder,ls='')  
                elif mode=='hatch':ax.fill([bd_int_loc[0],bd_int_loc[1],bd_int_loc[1],bd_int_loc[0]],[y_range_loc[0],y_range_loc[0],y_range_loc[1],y_range_loc[1]], fill=False, hatch='\\',color=facecolor,zorder=zorder)         
  
    return None






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy as np
from scipy.optimize import newton
from copy import deepcopy
from ..ANTARESS_general.constant_data import G_usi,Mjup,Msun
from ..ANTARESS_general.utils import np_where1D,npint,stop,closest

##################################################################################################    
#%% Coordinates
################################################################################################## 

def coord_expos(pl_loc,coord_dic,inst,vis,star_params,pl_params,bjd_inst,exp_time,data_dic,RpRs):
    r"""**Exposure coordinates**

    Calculates temporal and spatial coordinates of planets at the start, mid, and end of an exposure.
    
    Args:
        TBD
    
    Returns:
        None
    
    """ 
    
    #Orbital phase and corresponding times (days) of current exposure relative to transit center  
    st_phases,phases,end_phases=get_timeorbit(pl_loc,coord_dic[inst][vis],bjd_inst, pl_params, exp_time)[0:3]   

    #Exposure duration in phase
    ph_dur=end_phases-st_phases

    #Return values for start, mid, end exposure    
    xp_all,yp_all,zp_all,_,_,_,_,_,ecl= calc_pl_coord(pl_params['ecc'],pl_params['omega_rad'],pl_params['aRs'],pl_params['inclin_rad'],np.vstack((st_phases,phases,end_phases)),RpRs,pl_params['lambda_rad'],star_params)
    eclipse = ecl[0]
    st_positions= [xp_all[0][0],yp_all[0][0],zp_all[0][0]]
    positions= [xp_all[1][0],yp_all[1][0],zp_all[1][0]]
    end_positions= [xp_all[2][0],yp_all[2][0],zp_all[2][0]]

    #Calculation of planet orbital radial velocity in the star rest frame (km/s)
    if data_dic['Atm']['exc_plrange']:
  
        #Number of oversampling RV bins in exposure
        dph_loc=end_phases-st_phases
        nt_osamp_RV=max(int(dph_loc/data_dic['Atm']['dph_osamp_RVpl']),2)

        #Adjusted resolution
        dph_osamp_loc=dph_loc/nt_osamp_RV
        
        #Table of oversampling bins 
        #    - oversampling bins are centered in phase
        ph_osamp_loc = st_phases+dph_osamp_loc*(0.5+np.arange(nt_osamp_RV))  

        #Average orbital velocity over oversampled values
        #    - converted from velocity along the LOS in /Rstar/h to km/s
        rv_pl_osamp,v_pl_osamp = calc_pl_coord(pl_params['ecc'],pl_params['omega_rad'],pl_params['aRs'],pl_params['inclin_rad'],ph_osamp_loc,None,None,None,rv_LOS=True,omega_p=pl_params['omega_p'])[6:8]
        rv_pl_all=star_params['RV_conv_fact']*np.mean(rv_pl_osamp)    
        v_pl_all=star_params['RV_conv_fact']*np.mean(v_pl_osamp)   
                
    else:
        rv_pl_all=None
        v_pl_all=None
        
    return positions,st_positions,end_positions,eclipse,rv_pl_all,v_pl_all,st_phases,phases,end_phases,ph_dur

##################################################################################################    
#%%% Spatial and velocity coordinates
################################################################################################## 

def calc_Kstar(params,star_params):   
    r"""**Keplerian semi-amplitude.** 
    
    Calculates Keplerian semi-amplitude of the star induced by a given planet (in m/s).

    The general formula is
    
    .. math:: 
       K_\mathrm{prim} = (2 \pi G/P)^{1/3} M_\mathrm{sec} \sin(i_p) ((M_\mathrm{sec}+M_\mathrm{prim})^{1/3} / M_\mathrm{prim}) /\sqrt{1-e^2}
    
    With :math:`K_\mathrm{prim}` in km/s, `G` in :math:`m^3 kg^{-1} s^{-2}`, `P` in s, :math:`M_\mathrm{prim}` and :math:`M_\mathrm{sec}` in kg. Thus    

    .. math:: 
       K_\mathrm{prim} &= (212.91918458020422/P_\mathrm{days}^{1/3}) M_\mathrm{sec} \sin(i_p) ((M_\mathrm{sec}+M_\mathrm{prim})^{1/3}/M_\mathrm{prim}) / \sqrt{1-e^2}   \\
       &\mathrm{where\,} [ (2 \pi 6.67300\times10^{-11})^{1/3} 1.9891001\times10^{30} (1.9891001\times10^{30})^{1/3}]/[(24 \times 3600)^{1/3} (1.9891001\times10^{30})]/1000 = 212.91918458020422

    With :math:`P_\mathrm{days}` in days, :math:`M_\mathrm{prim}` and :math:`M_\mathrm{sec}` in :math:`M_\mathrm{Sun}`.  
    For a planet :math:`M_\mathrm{sec} = M_p << M_\mathrm{prim} = M_{\star}`, thus
    
    .. math::     
       K_{\star} \sim (2 \pi G/P)^{1/3} (M_p \sin(i_p)/M_{\star}^{2/3}) /\sqrt{1-e^2}

    With :math:`K_{\star}` in km/s, `G` in :math:`m^3 kg^{-1} s^{-2}`, `P` in s, :math:`M_p` and :math:`M_{\star}` in kg. Or
    
    .. math:: 
       K_{\star} &= (0.20323137323712528/P_\mathrm{days}^{1/3}) (M_p \sin(i_p)/M_{\star}^{2/3}) /\sqrt{1-e^2} \\
       &\mathrm{where\,} [ (2 \pi 6.67300\times10^{-11})^{1/3} 1.8986000\times10^{27}]/[(24 \times 3600)^{1/3} (1.9891001\times10^{30})^{2/3}]/1000 = 0.20323137    

    With :math:`P_\mathrm{days}` in days, :math:`M_p` and :math:`M_{\star}` in :math:`M_\mathrm{Sun}`. Or

    .. math:: 
       K_{\star} = (0.02843112300449059/P_\mathrm{years}^{1/3}) (M_p \sin(i_p)/M_{\star}^{2/3}) /\sqrt{1-e^2}
 
    With :math:`P_\mathrm{years}` in years, :math:`M_p` and :math:`M_{\star}` in :math:`M_\mathrm{Sun}`.   

    Args:
        TBD
    
    Returns:
        None
    
    """     
    return (2.*np.pi*G_usi/(params['period_s']))**(1./3.)*(params['Msini']*Mjup/(star_params['Mstar']*Msun)**(2./3.))*1./np.sqrt(1.-params['ecc']**2.)


def calc_rv_star(coord_dic,inst,vis,system_param,gen_dic,bjd_exp,dur_exp,sysvel):
    r"""**Stellar rv: for observations**

    Calculates the rv of the star in the stellar and sun barycentric rest frames.

    The Keplerian rv of the star induced by a given planet is
    
    .. math::     
       rv(\mathrm{star/stellar \, CDM \,;\, from \, pl})= K_{\star} ( \cos(\nu+\omega_\mathrm{bar})+ e \cos(\omega_\mathrm{bar}) )
       
    With :math:`K_{\star}`, :math:`\nu`, `e` and :math:`\omega_\mathrm{bar}` specific to the planet.
    Radial velocity is negative when the planet is coming toward the observer, thus in the same frame as the observed rv tables.

    Args:
        TBD
    
    Returns:
        None
    
    """     

    #Oversampling factor of exposures
    n_ov=10.

    #-------------------------------------    
    #Keplerian motion induced by requested planets
    RV_star_stelCDM_ov=0.
    for pl_loc in gen_dic['kepl_pl']:     
        PlParam_loc=system_param[pl_loc]   
        
        #Orbital phase 
        st_phase,phase,end_phase=get_timeorbit(pl_loc,coord_dic[inst][vis],bjd_exp, PlParam_loc, dur_exp)[0:3]

        #True anomaly for start and end of exposure
        True_anom=calc_true_anom(PlParam_loc['ecc'],np.vstack((st_phase,end_phase)),PlParam_loc['omega_rad'])[0]

        #Oversampling of true anomaly over exposure
        min_TA=np.min(True_anom)
        max_TA=np.max(True_anom)
        
        #Oversampling
        dTrueAnom_ov=(max_TA-min_TA)/n_ov            
        True_anom_ov=min_TA+dTrueAnom_ov*(0.5+np.arange(n_ov,dtype=float))

        #Keplerian motion induced by the planet (km/s)
        RV_star_stelCDM_ov+=PlParam_loc['Kstar_kms']*(np.cos(True_anom_ov+PlParam_loc['omega_rad'])+PlParam_loc['ecc']*np.cos(PlParam_loc['omega_rad']) )                
    
    #-------------------------------------    

    #Average oversampled RV (km/s)
    RV_star_stelCDM=np.mean(RV_star_stelCDM_ov)  
    
    #Stellar radial velocity relative to the sun
    RV_star_solCDM=RV_star_stelCDM+sysvel

    return RV_star_stelCDM,RV_star_solCDM






def calc_rv_star_HR(pl_ref,system_param,kepl_pl,coord_dic,inst,vis,data_dic):
    r"""**Stellar rv: for plots**

    Calculates the rv of the star in the sun barycentric rest frame, over a high-resolution phase table.
    
    Specific to each visit, since the Keplerian rv is not the same if there are multiple planets.

    Args:
        TBD
    
    Returns:
        None
    
    """ 
    
    #Phase resolution and high-resolution phase table
    #    - we extend slightly the table on both sides
    #    - we take a temporal resolution shorter than the shorter exposure in the visit
    nexp_HR=10
    t_dur_d=coord_dic[inst][vis]['t_dur_d']
    min_texpd=np.min(t_dur_d)
    dbjd_HR=min_texpd/nexp_HR
    min_bjd=np.min(coord_dic[inst][vis]['bjd']-t_dur_d)
    max_bjd=np.max(coord_dic[inst][vis]['bjd']+t_dur_d)
    nbjd_HR=int((max_bjd-min_bjd)/dbjd_HR)+1
    bjd_tab_HR=min_bjd+dbjd_HR*np.arange(nbjd_HR)   
    phase_tab_HR=np.zeros(nbjd_HR, dtype=float)
    if pl_ref not in kepl_pl:stop('Current planet not accounted for in Keplerian motion')
    
    #Radial velocity table
    RV_star_stelCDM = np.zeros(nbjd_HR, dtype=float)

    #Motion for all instrument exposures
    for i_loc,bjd_loc in enumerate(bjd_tab_HR):
        for ipl,pl_loc in enumerate(kepl_pl):
            PlParam=system_param[pl_loc]
                
            #Orbital phase 
            phase_temp=(bjd_loc - (coord_dic[inst][vis][pl_loc]['Tcenter'] - 2400000.))/PlParam["period"]
            phase = (phase_temp - npint(phase_temp+np.sign(phase_temp)*0.5)) 
            if pl_loc==pl_ref:phase_tab_HR[i_loc]=phase            
            
            #True anomaly
            ecc=PlParam['ecc']
            omega_bar=PlParam['omega_rad']          
            True_anom,Ecc_anom,_=calc_true_anom(ecc,phase,omega_bar)
  
            #Keplerian motion (km/s)
            RV_star_stelCDM[i_loc]+=PlParam['Kstar_kms']*(np.cos(True_anom+omega_bar)+ecc*np.cos(omega_bar) )

    ### End of loop on exposures 
 
    #Keplerian curve in heliocentric rest frame (km/s)
    RV_star_solCDM=RV_star_stelCDM+data_dic['DI']['sysvel'][inst][vis]

    return phase_tab_HR,RV_star_solCDM




def Kepler_func(Ecc_anom,Mean_anom,ecc):
    r"""**Eccentric anomaly solver.**
    
    Finds the eccentric anomaly of the orbit using the mean anomaly and eccentricity.
    
    :math:`M = E - e \sin(E)`        
    
    Args:
        Ecc_anom (float): Eccentric anomaly.
        Mean_anom (float): Mean anomaly.
        ecc (float): Eccentricity.
    
    Returns:
        delta (float): equation to nullify 
    
    """     
    delta=Ecc_anom-ecc*np.sin(Ecc_anom)-Mean_anom
    return delta 



   
def calc_mean_anom_TR(ecc,omega_bar):  
    r"""**Mean anomany at mid-transit.**
    
    Args:
        ecc (float): Eccentricity.
        omega_bar (float): Argument of periastron.
    
    Returns:
        Mean_anom_TR (float): Mean anomany at mid-transit. 
    
    """     
    
    #True anomaly of the planet at mid-transit (in rad):
    #    - angle counted from 0 at the periastron, to the star/Earth LOS
    #    - >0 counterclockwise, possibly modulo 2pi
    #    - with omega_bar counted from ascending node to periastron
    True_anom_TR=(np.pi/2.)-omega_bar

    #Mean anomaly at the time of the transit
    #    - corresponds to 'dt_transit' (in years), time from periapsis to transit center 
    #    - atan(X) is in -pi/2 ; pi/2 
    Ecc_anom_TR=2.*np.arctan( np.tan(True_anom_TR/2.)*np.sqrt((1.-ecc)/(1.+ecc)) )
    Mean_anom_TR=Ecc_anom_TR-ecc*np.sin(Ecc_anom_TR)
    if (Mean_anom_TR<0.):Mean_anom_TR=Mean_anom_TR+2.*np.pi    
    
    return Mean_anom_TR
    


def calc_true_anom(ecc,phase,omega_bar):
    r"""**True anomaly.**
    
    Args:
        ecc (float): Eccentricity.
        phase (1D array): orbital phase.
        omega_bar (float): Argument of periastron.
    
    Returns:
        True_anom (1D array): True anomany at given phases. 
        Ecc_anom (1D array): Eccentric anomany at given phases. 
        Mean_anom (1D array): Mean anomany at given phases. 
    
    """  
    
    #Circular orbit
    #  - 1e-4 instead of 0 to avoid loosing too much time with really low tested values of ecc    
    if (ecc<1e-4):

        #Mean anomaly
        Mean_anom=2.*np.pi*phase

        #Eccentric anomaly
        Ecc_anom=Mean_anom
        
        #True anomaly
        True_anom=Mean_anom
                                
    #--------------------------------------------------------------------------------------------------------------------
    #Eccentric orbit   
    #  - by definition the ascending node is where the object goes toward the observer through the plane of sky
    #  - omega_bar is the angle between the ascending node and the periastron, in the orbital plane (>0 counterclockwise)                       
    else:

        #Mean anomaly
        #  - time origin of t_mean at the periapsis (t_mean=0 <-> M=0 <-> E=0)
        #  - M(t_mean)=M(dt_transit)+M(t_simu) 
        Mean_anom=2.*np.pi*phase+calc_mean_anom_TR(ecc,omega_bar)
			        
        #Eccentric anomaly : 
        #  - M = E - e sin(E)
        #    - >0 counterclockwise
        #  - angle, with origin at the ellipse center, between the major axis toward the periapsis and the 
        #line crossing the circle with radius 'aRs' at its intersection with the perpendicular to
        #the major axis through the planet position
        if np.isscalar(Mean_anom):
            Ecc_anom=newton(Kepler_func,Mean_anom,args=(Mean_anom,ecc,))
        else:
            dim_input = np.shape(Mean_anom)
            if len(dim_input)==1:Ecc_anom=np.array([newton(Kepler_func,Mean_anom_loc,args=(Mean_anom_loc,ecc,)) for Mean_anom_loc in Mean_anom])
            else:
                Ecc_anom = np.zeros(dim_input,dtype=float)
                for idim in range(dim_input[0]):Ecc_anom[idim]=np.array([newton(Kepler_func,Mean_anom_loc,args=(Mean_anom_loc,ecc,)) for Mean_anom_loc in Mean_anom[idim]])

        #True anomaly of the planet at current time
        True_anom=2.*np.arctan(np.sqrt((1.+ecc)/(1.-ecc))*np.tan(Ecc_anom/2.))
                
    return True_anom,Ecc_anom,Mean_anom 

 

  



def calc_pl_coord(ecc,omega_bar,aRs,inclin,ph_loc,RpRs,lambda_rad,star_params,rv_LOS=False,omega_p=None):
    r"""**Planet orbit**

    Calculates orbital planetary position and velocity, as well as eclipse condition.
    
    References:

        - `<http://www.gehirn-und-geist.de/sixcms/media.php/370/Leseprobe.406372.pdf>`_
        - `<http://www.relativitycalculator.com/pdfs/RV_Derivation.pdf>`_
        - `<https://www.sciencedirect.com/topics/engineering/eccentric-anomaly>`_

    We calculate the perifocal coordinates of the planet, in cartesian coordinates relative to the star.
    The radial velocity :math:`rv(pl/star)` (negative when the planet is coming toward us) is derived directly from these coordinates as

    .. math::      
       rv(pl/star) = - K_\mathrm{pl/star} ( \cos(\nu+\omega_\mathrm{bar})+e \cos(\omega_\mathrm{bar}) )    

    Kepler third's law gives 
     
    .. math::       
       P^2/a^3 &= 4 \pi^2 / (G (M_\mathrm{pl}+M_{\star}))    \\
       a &= ( G P^2 (M_{\star}+M_\mathrm{pl}) / (4 \pi^2) )^{1/3}
          
    Thus we can write the semi-amplitude of the planet Keplerian motion as
          
    .. math::      
       K_\mathrm{pl/star} &= 2 \pi a \sin(i_p)/(P \sqrt{1-e^2})    \\   
       K_\mathrm{pl/star} &= (2 \pi G/P)^{1/3} (M_{\star}+M_\mathrm{pl})^{1/3} \sin(i_p)/\sqrt{1-e^2}        
    
    The radial velocity of the planet in the star rest frame can also be defined as 

    .. math::  
       rv(pl/star) = rv(pl/CMD_{\star}) - rv(star/CMD_{\star})
    
    And because the barycenter is fixed in its own frame:

    .. math::          
       & M_{\star} rv(star/CMD_{\star}) +  M_\mathrm{pl} rv(pl/CMD_{\star}) = 0          \\  
       & M_{\star} rv(star/CMD_{\star}) +  M_\mathrm{pl} (rv(pl/star)+rv(star/CMD_{\star})) = 0       \\     
       & rv(star/CMD_{\star})= - rv(pl/star) M_\mathrm{pl}/(M_{\star}+M_\mathrm{pl}) 
     
    Thus 

    .. math:: 
       rv(star/CMD_{\star}) &= K[star/CMD_{\star}] ( \cos(\nu+\omega_\mathrm{bar})+e \cos(\omega_\mathrm{bar}) )     \\
       \mathrm{where\,} & K[star/CMD_{\star}] = K_\mathrm{pl/star} M_\mathrm{pl}/(M_{\star}+M_\mathrm{pl}) \\
                        & K[star/CMD_{\star}] = (2 \pi G/P)^{1/3} M_\mathrm{pl} \sin(i_p)/((M_{\star}+M_\mathrm{pl})^{2/3} \sqrt{1-e^2}) 
    
    With `K` in km/s, `G` in :math:`m^3 kg^{-1} s^{-2}`, `P` in s, :math:`M_\mathrm{pl}` and :math:`M_{\star}` in kg. Or (see `calc_Kstar()`)  

    .. math::  
       K[star/CMD_{\star}] &= (212.91918458020422/P_\mathrm{days}^{1/3}) M_\mathrm{pl,Msun} \sin(i_p)/((M_{\star,Msun} +M_\mathrm{pl,Msun})^{2/3} \sqrt{1-e^2}) \\
       K[star/CMD_{\star}] &= (0.02843112300449059/P_\mathrm{yr}^{1/3}) M_\mathrm{pl,Msun} \sin(i_p)/((M_{\star,Msun} +M_\mathrm{pl,Msun})^{2/3} \sqrt{1-e^2})           
              
    We can also write  

    .. math:: 
       rv(pl/CMD_{\star}) &=  rv(pl/star) + rv(star/CMD_{\star})  \\
                          &=  ( - K_\mathrm{pl/star} + K_\mathrm{pl/star} M_\mathrm{pl}/(M_{\star}+M_\mathrm{pl}) ) ( \cos(\nu+\omega_\mathrm{bar})+e \cos(\omega_\mathrm{bar}) )     \\  
                          &=  K[pl/CMD_{\star}] ( \cos(\nu+\omega_\mathrm{bar})+e \cos(\omega_\mathrm{bar}) )  \\
       \mathrm{where\,} & K[pl/CMD_{\star}] = K_\mathrm{pl/star} ( M_\mathrm{pl}/(M_{\star}+M_\mathrm{pl})  - 1 )  \\
                        & K[pl/CMD_{\star}] = K_\mathrm{pl/star} ( (M_\mathrm{pl}-M_{\star})/(M_{\star}+M_\mathrm{pl}) )  \\
                        & K[pl/CMD_{\star}] = - (2 \pi G/P)^{1/3} (M_{\star}-M_\mathrm{pl}) \sin(i_p)/(\sqrt{1-e^2} (M_{\star}+M_\mathrm{pl})^{1/3} )

    If we assume that the atmospheric signal tracks the planet orbital velocity (ie, there is no atmospheric dynamics), then we can 
    link the planet and star masses with this measurement. Doing as few assumption as possible, the quantity we measure is 
     
    .. math::     
       rv(pl/CDM_\mathrm{sun}) &= rv(pl/CMD_{\star}) + rv(CMD_{\star}/CDM_\mathrm{sun})  \\
       rv(pl/CDM_\mathrm{sun}) &= - (2 \pi G/P)^{1/3} (M_{\star}-M_\mathrm{pl}) \sin(i_p)/(\sqrt{1-e^2} (M_{\star}+M_\mathrm{pl})^{1/3} ) + rv_\mathrm{sys}   
     
    Args:
        TBD
    
    Returns:
        None
    
    """     
    c_ip = np.cos(inclin)       
    s_ip = np.sin(inclin)       
      
    #True anomaly
    True_anom_loc,Ecc_anom_loc,_=calc_true_anom(ecc,ph_loc,omega_bar)  

    #Orbital frequency (yr-1 to h-1) and velocity coordinates initialization 
    if rv_LOS:omega_p_h = omega_p/(365.2425*24.) 
    else:vxp_loc,vyp_loc,vzp_loc=None,None,None
            
    #Circular orbit
    #  - 1e-4 instead of 0 to avoid loosing too much time with really low tested values of ecc    
    if (ecc<1e-4):

        #Coordinates in the orbital plane oriented toward Earth
        #    - axis:
        # + X1 is the node line, in the plane of sky (ie the plane perpendicular to the LOS)
        #      obtained by rotation of Z1 in the direction of the orbital motion 
        # + Y1 is the perpendicular to the node line in the orbital plane (oriented toward the half space containing Earth)
        # + Z1 is the normal to the orbital plane   
        #      z1 is null because the planet moves in the x1,y1 plane 
        #    - coordinates in /Rstar and /Rstar/h
        X1_p = aRs*np.sin(True_anom_loc)        
        Y1_p = aRs*np.cos(True_anom_loc)
        if rv_LOS:      
            VX1_p= omega_p_h*Y1_p    
            VY1_p=-omega_p_h*X1_p

        #Coordinates in the sky-projected stellar referential 
        #    - axis:
        # + X2 = X1         
        # + Y2 = Z1 obtained as Z2
        #      projection of the oriented normal (ie the spin axis of the planet motion along its orbit) to the orbital plane in the plane of sky (perpendicular to the LOS)  
        #      the planet always moves in front of the star from J2- to J2+ (with I2+ toward the Earth)
        #      y2 = - y1 cos(inclin) + z1 sin(inclin)                
        # + Z2 is the line of sight from star center toward Earth
        #      obtained by rotation of Y1 around X1 by inclin-90 (inclin = angle from the LOS to the normal of the orbital plane)
        #      z2 = y1 sin(inclin) + z1 cos(inclin)   
        #    - coordinates in /Rstar and /Rstar/h
        #for a circular orbit, only the inclination is necessary, since the orbital plane is oriented so that the
        #node line is the same as the Y axis of the stellar referential 
        xp_loc =  X1_p
        yp_loc = -Y1_p*c_ip
        zp_loc =  Y1_p*s_ip
        if rv_LOS:
            vxp_loc=  VX1_p 
            vyp_loc= -VY1_p*c_ip              
            vzp_loc=  VY1_p*s_ip          
            
            
    #--------------------------------------------------------------------------------------------------------------------
    #Eccentric orbit   
    #  - by definition the ascending node is where the object goes toward the observer through the plane of sky
    #  - omega_bar is the angle between the ascending node and the periastron, in the orbital plane (>0 counterclockwise)                       
    else:

        #Planet position in the orbital plane in Rstar
        #  - origin at the focus (star)
        #  - I0 : major axis ( a*cos(E) - a*cos(e) ) oriented toward the periapsis
        #  - J0 : perpendicular to I0 ( b*sin(E) ) when rotating with the orbital motion
        #  - K0 : normal to the orbital plane 
        c_ecc = np.cos(Ecc_anom_loc)
        s_ecc = np.sin(Ecc_anom_loc)        
        X0_p=aRs*(c_ecc-ecc)
        Y0_p=aRs*np.sqrt(1.-ecc*ecc)*s_ecc
    
        #Coordinates in the orbital plane oriented toward the Earth in Rstar
        #Omega_bar : angle between the ascending node and the periastron in the orbital plane
        #  - I1 is the node line (from K1 when rotating with the orbital motion)    
        #    Y1 is the perpendicular to the node line in the orbital plane, toward the Earth
        #  - Z1 = ( sin(omega_bar) # cos(omega_bar)  ) in (I0,J0)
        #    Y1 = ( -cos(omega_bar) # sin(omega_bar)  ) in (I0,J0)
        c_om = np.cos(omega_bar)
        s_om = np.sin(omega_bar)
        X1_p =  -X0_p*c_om + Y0_p*s_om
        Y1_p =   X0_p*s_om + Y0_p*c_om    

        #Coordinates in the plane containing the LOS and the node line
        #   - Z : LOS toward Earth (from I1 by rotating with angle inclin)
        #   - X : node line (=K1)          
        #   - Y : complete the right handed referential
        xp_loc=X1_p            
        yp_loc=-Y1_p*c_ip
        zp_loc= Y1_p*s_ip

        #Velocity coordinates
        #    - in /Rstar/h 
        if rv_LOS:

            #Distance star - planet (/Rstar)
            Dplanet=aRs*(1.-ecc*c_ecc)
 
            #Planet velocity in the orbital plane 
            #  - vx = - a sin(E) dE/dt
            #    vy = a sqrt(1-e^2) cos(E) dE/dt
            #  - dM = omega_p_h*dt = dE ( 1 - e cos(E) ) = dE*Dplanet/a_pl  
            VX0_p =-omega_p_h*(pow(aRs,2.))*s_ecc/Dplanet
            VY0_p = omega_p_h*(pow(aRs,2.))*np.sqrt(1-pow(ecc,2.))*c_ecc/Dplanet    

            #Planet velocity in the orbital plane oriented toward the Earth   
            VX1_p = -VX0_p*c_om + VY0_p*s_om  
            VY1_p =  VX0_p*s_om + VY0_p*c_om
            
            #Coordinates in the sky-projected referential 
            vxp_loc=  VX1_p
            vyp_loc= -VY1_p*c_ip 
            vzp_loc=  VY1_p*s_ip   

    #--------------------------------------------------------------------------------------------------------------------
    #Absolute orbital velocity
    #    - in /Rstar/h 
    if rv_LOS:vp_loc=np.sqrt(VX1_p*VX1_p + VY1_p*VY1_p)
    else:vp_loc=None

    #--------------------------------------------------------------------------------------------------------------------
    #Distance star - planet in the plane of sky
    #    - in /Rstar
    Dprojplanet=np.sqrt(xp_loc*xp_loc + yp_loc*yp_loc)

    #--------------------------------------------------------------------------------------------------------------------
    #Eclipse status  
    if RpRs is not None:ecl_loc = eclipse_status(Dprojplanet,RpRs,lambda_rad,star_params,xp_loc,yp_loc,zp_loc)     
    else:ecl_loc = None
          
    return xp_loc,yp_loc,zp_loc,Dprojplanet,vxp_loc,vyp_loc,vzp_loc , vp_loc, ecl_loc   
    
    
def calc_pl_coord_plots(n_pts_orbit,pl_params):  
    r"""**Planet orbit: for plots.**
    
    Defines planetary orbit coordinates for plot and contact times.
    Coordinates are defined in the sky-projected orbital frame
        
    .. math:: 
       X_\mathrm{sky} &= \mathrm{node \, line \, of \, orbital \, plane}  \\
       Y_\mathrm{sky} &= \mathrm{projection \, of \, the \, orbital \, plane \, normal} \\
       Z_\mathrm{sky} &= \mathrm{LOS}

    Args:
        TBD
    
    Returns:
        coord_orbit (array): sky-projected coordinates. 
    
    """    
    
    #Orbital properties
    #    - Inclin is the inclination from the LOS to the normal to the orbital plane
    ecc=pl_params['ecc']
    aRs=pl_params['aRs']
    Inclin=pl_params['inclin_rad']
    c_ip = np.cos(Inclin)
    s_ip = np.sin(Inclin)
    omega_bar=pl_params['omega_rad']
 
    #Circular orbit  
    if (ecc<1e-4):    
        n_pts_orbit=int(n_pts_orbit)
        ph_plot=np.arange(n_pts_orbit+1.)/n_pts_orbit
        X0_plot= aRs*np.cos(2.*np.pi*ph_plot)
        Y0_plot= aRs*np.sin(2.*np.pi*ph_plot)
        coord_orbit=[Y0_plot,              #Xsky
                     -X0_plot*c_ip,        #Ysky 
                      X0_plot*s_ip]        #Zsky   
  
    #Elliptic orbit 
    else:
        
        #Time resolution of the orbit
        #    - we set a high resolution around the periastron   
        #    - see planet_coord for details  
        n_pts_horbit=int(n_pts_orbit/2.)
        ph_plot=-(0.05)+(0.1/n_pts_horbit)*np.arange(n_pts_horbit+1)
        ph_plot=np.append(ph_plot,(0.05)+((1-0.1 )/n_pts_horbit)*np.arange(n_pts_horbit+1))
        Mean_anom_plot=2.*np.pi*ph_plot
        Ecc_anom_plot=np.array([newton(Kepler_func,Mean_anom_plot[i],args=(Mean_anom_plot[i],ecc,)) for i in range(len(Mean_anom_plot))])
    
        #Coord. in the orbit plane with semi-major axis as X axis    
        X0_plot=aRs*(np.cos(Ecc_anom_plot)-ecc)
        Y0_plot=aRs*np.sqrt(1.-pow(ecc,2.))*np.sin(Ecc_anom_plot)

        #Coord. in the XYZ simulation referential     
        coord_orbit = [-X0_plot*np.cos(omega_bar) +  Y0_plot*np.sin(omega_bar),                  #Xsky
                      -( X0_plot*np.sin(omega_bar) +  Y0_plot*np.cos(omega_bar) )*c_ip,   #Ysky
                       ( X0_plot*np.sin(omega_bar) +  Y0_plot*np.cos(omega_bar) )*s_ip]   #Zsky
    
    return coord_orbit    
    


def calc_zLOS_oblate(x_st_sk,y_st_sk,istar_rad,RpoleReq):
    r"""**LOS coordinate for oblate photosphere.**

    Calculates the `Z` coordinate (along the LOS) of an element at `(X,Y)` over the sky-projected photosphere (inclined star frame) of an oblate star. 
    The `Z` coordinate is returned in the stellar hemispheres facing and opposite the observer.

    Based on Barnes+2009, Eq. 13 and 14 with :math:`\Phi = \pi/2 - i_{\star}`. 
    Beware that their Eq 14 is wrong, as :math:`1-f^2` should be :math:`(1-f)^2`

    Args:
        TBD
    
    Returns:
        None
    
    """    
    
    #The photosphere is defined in the star frame as (Eq 12 in Barnes 2009):
    #xstar^2 + ystar^2 / Rpole^2 + zstar^2 = 1
    #in coordinates of the sky-projected star frame:
    #xstar = x
    #ystar = y*si + z*ci
    #zstar =-y*ci + z*si
    #x^2 + (y*si + z*ci)^2 / Rpole^2 + (-y*ci + z*si)^2 = 1
    #(x*Rpole)^2 + y^2*si^2 + 2*y*z*ci*si + z^2*ci^2 + y^2*ci^2*Rpole^2 - 2*y*ci*z*si*Rpole^2 + z^2*si^2*Rpole^2 = Rpole^2
    #z^2*ci^2 + z^2*si^2*Rpole^2    +  2*y*z*ci*si - 2*y*ci*z*si*Rpole^2   + (x*Rpole)^2 + y^2*si^2 + y^2*ci^2*Rpole^2 - Rpole^2 = 0
    #z^2*(ci^2 + si^2*Rpole^2) +  2*y*z*ci*si*(1 - Rpole^2) + y^2*(si^2 + ci^2*Rpole^2) + Rpole^2*(x^2 - 1) = 0
    #z^2*(1 - si^2*(1-Rpole^2)) +  2*y*z*ci*si*(1 - Rpole^2) + y^2*si^2*(1- Rpole^2) + Rpole^2*(x^2+ y^2 - 1) = 0
    #   A = (1 - si^2*(1-Rpole^2))
    #   B = 2*y*ci*si*(1 - Rpole^2)
    #   C = y^2*si^2*(1- Rpole^2) + Rpole^2*(x^2+ y^2 - 1)
    #delta = B^2 - 4*A*C
    #z = (-B+-sqrt(delta))/(2*A)
    #there is no Req in the equation because coordinates are all normalized by Rstar
    ci = np.cos(istar_rad)
    si = np.sin(istar_rad)
    mRp2 = (1. - RpoleReq**2.)
    Aquad = 1. - si**2.*mRp2
    Bquad = 2.*y_st_sk*ci*si*mRp2
    Cquad = y_st_sk**2.*si**2.*mRp2 + RpoleReq**2.*(x_st_sk**2.+ y_st_sk**2. - 1.) 
    det = Bquad**2.-4.*Aquad*Cquad
    cond_def = det>=0.
    z_st_sky_behind = np.zeros(len(x_st_sk),dtype=float)*np.nan
    z_st_sky_front = np.zeros(len(x_st_sk),dtype=float)*np.nan
    if True in cond_def:
        z_st_sky_behind[cond_def] = (-Bquad[cond_def]-np.sqrt(det[cond_def]))/(2.*Aquad)
        z_st_sky_front[cond_def] = (-Bquad[cond_def]+np.sqrt(det[cond_def]))/(2.*Aquad)
    return z_st_sky_behind,z_st_sky_front,cond_def



def excl_plrange(cond_def,range_star_in,iexp,edge_bins,data_type):
    r"""**Planet atmospheric masking.**

    Identifies spectral pixels contaminated by the planetary atmosphere, as requested in input. 
    A common rv range is excluded for each spectral line in the requested planet atmosphere mask 
    (see `init_visit()` for the definition of these ranges in other frames).

    Args:
        TBD
    
    Returns:
        None
    
    """ 
    cond_kept = np.ones(cond_def.shape,dtype=bool)
    idx_excl_bd_ranges = []
    if data_type=='CCF':
        range_star = range_star_in['CCF']
        for pl_loc in range_star:
            idx_excl = np_where1D((edge_bins[0:-1]>=range_star[pl_loc][0,iexp]) & (edge_bins[1:]<=range_star[pl_loc][1,iexp]))
            if len(idx_excl)>0:
                idx_excl_bd_ranges+=[[idx_excl[0],idx_excl[-1]]]
                cond_kept[idx_excl] = False 

    elif 'spec' in data_type:
        range_star = range_star_in['spec']
        
        #Defined bins in spectrum
        #    - we do not check pixels already undefined to calibration time
        idx_def_exp_pl = np_where1D(cond_def)                             
        n_idx_def_loc = len(idx_def_exp_pl)
        
        #Process lines that overlap at least partially with spectrum range
        #    - we use the spectral table of the target exposure, since current exposure has either been resampled over it or share the same table 
        for pl_loc in range_star:
            cond_keepL = (range_star[pl_loc][:,1,iexp]>=edge_bins[idx_def_exp_pl[0]]) &  (range_star[pl_loc][:,0,iexp]<=edge_bins[idx_def_exp_pl[-1]+1])
            if np.sum(cond_keepL)>0:

                #Indexes of first and last defined bins in the exposure table that overlap with the line range
                #    - see method in resample_func()
                idx_first_overpix =np.searchsorted(edge_bins[idx_def_exp_pl],range_star[pl_loc][cond_keepL,0,iexp],side='right')-1
                idx_last_overpix =np.searchsorted(edge_bins[idx_def_exp_pl+1],range_star[pl_loc][cond_keepL,1,iexp],side='left')
                idx_first_overpix[idx_first_overpix==-1]=0        
                idx_last_overpix[idx_last_overpix==n_idx_def_loc]=n_idx_def_loc-1         
        
                #Exclude progressively the velocity range relative to each line 
                #    - bins are relative to idx_def_exp_pl
                for idx_first_overpix_loc,idx_last_overpix_loc in zip(idx_first_overpix,idx_last_overpix):
                    idx_undef_sub = np.arange(idx_first_overpix_loc,idx_last_overpix_loc+1)
                    idx_undef = idx_def_exp_pl[idx_undef_sub]
                    idx_excl_bd_ranges+=[[idx_undef[0],idx_undef[-1]]]
                    cond_kept[idx_undef] = False
        
    return cond_kept,idx_excl_bd_ranges






##################################################################################################    
#%%% Time coordinates
################################################################################################## 
  
def get_timeorbit(pl_loc,coord_vis,bjd_tab, PlParam, exp_time_tab):
    r"""**Exposure time and phase**

    Calculates time and phase coordinates of planets at the start, mid, and end of an exposure.

    Args:
        TBD
    
    Returns:
        None
    
    """ 
    
    #Transit center (days)
    Tcen= coord_vis[pl_loc]['Tcenter'] - 2400000.  

    #Orbital phase (between +-0.5) and corresponding times (d)
    phase_temp=(bjd_tab-Tcen)/PlParam["period"]
    mid_phase_tab = (phase_temp - npint(phase_temp+np.sign(phase_temp)*0.5))
    mid_time_tab  = mid_phase_tab*PlParam["period"]
    if exp_time_tab is not None:
        
        #Half exposure time (days)
        texp_d_tab=exp_time_tab/(2.*24.*3600.)

        #Times
        phase_temp=(bjd_tab-texp_d_tab-Tcen)/PlParam["period"]
        st_phase_tab = (phase_temp - npint(phase_temp+np.sign(phase_temp)*0.5))
        st_time_tab  = st_phase_tab*PlParam["period"]
        phase_temp=(bjd_tab+texp_d_tab-Tcen)/PlParam["period"]
        end_phase_tab = (phase_temp - npint(phase_temp+np.sign(phase_temp)*0.5))
        end_time_tab  = end_phase_tab*PlParam["period"]
        
    else:st_phase_tab,end_phase_tab,st_time_tab,end_time_tab=None,None,None,None

    return st_phase_tab,mid_phase_tab,end_phase_tab,st_time_tab,mid_time_tab,end_time_tab


def conv_phase(coord_dic,inst,vis,system_param,ref_pl,pl_loc,phase_tab_in):
    r"""**Planet phase conversion**

    Converts orbital phase of reference planet for another planet.

    Args:
        TBD
    
    Returns:
        None
    
    """ 
    if (ref_pl is None) or (pl_loc==ref_pl):     
        phase_tab_out=phase_tab_in
    else:
        phase_tab_out = deepcopy(phase_tab_in) 
        Tcenter_ref = coord_dic[inst][vis][ref_pl]['Tcenter']
        Tcenter_loc = coord_dic[inst][vis][pl_loc]['Tcenter']   
        for iph in range(phase_tab_in.shape[0]):   

            #Absolute time table and corresponding phases   
            cen_bjd = Tcenter_ref+phase_tab_in[iph]*system_param[ref_pl]["period"]
            phase_temp=(cen_bjd-Tcenter_loc)/system_param[pl_loc]["period"]
            phase_tab_out[iph] = (phase_temp - npint(phase_temp+np.sign(phase_temp)*0.5))        

    return phase_tab_out




def calc_tr_contacts(RpRs,pl_params,stend_ph,star_params):  
    r"""**Transit contacts.**

    Calculates the orbital phase of the four transit contacts for the selected planet.

    Args:
        TBD
    
    Returns:
        None
    
    """     
    ecc=pl_params['ecc']
    aRs=pl_params['aRs']
    omega_bar=pl_params['omega_rad']
    Inclin=pl_params['inclin_rad']
    c_ip = np.cos(Inclin) 
    s_ip = np.sin(Inclin) 
    
    #Defining contacts numerically over a high-resolution phase table around the transit. 
    contact_phases=np.zeros(4,dtype=float)*np.nan
    n_pts_contacts=int(5000)   
    ph_ineg=np.arcsin(1./aRs)/(2.*np.pi)   #phase approximative de l'ingress
    ph_st=-stend_ph*ph_ineg
    ph_end=stend_ph*ph_ineg
    ph_contacts=ph_st+((ph_end-ph_st)/n_pts_contacts)*np.arange(n_pts_contacts+1.)
    
    if (ecc > 1e-4):
        if 'Mean_anom_TR' not in pl_params:pl_params['Mean_anom_TR'] = calc_mean_anom_TR(ecc,omega_bar) 
        Mean_anom_plot=2.*np.pi*ph_contacts+pl_params['Mean_anom_TR']
        Ecc_anom_plot=np.array([newton(Kepler_func,Mean_anom_plot[i],args=(Mean_anom_plot[i],ecc,)) for i in range(len(Mean_anom_plot))])   
        X0_plot=aRs*(np.cos(Ecc_anom_plot)-ecc)
        Y0_plot=aRs*np.sqrt(1.-pow(ecc,2.))*np.sin(Ecc_anom_plot)
        xp=-X0_plot*np.cos(omega_bar) +  Y0_plot*np.sin(omega_bar)
        yp=-( X0_plot*np.sin(omega_bar) +  Y0_plot*np.cos(omega_bar) )*c_ip
        zp=( X0_plot*np.sin(omega_bar) +  Y0_plot*np.cos(omega_bar) )*s_ip
    else:
        X0_plot= aRs*np.cos(2.*np.pi*ph_contacts)
        Y0_plot= aRs*np.sin(2.*np.pi*ph_contacts)
        xp=Y0_plot
        yp=-X0_plot*c_ip
        zp=X0_plot*s_ip
        
    #Points before transit, front of the star
    w_bef=np.where((xp<0) & (zp>0))[0]        

    #Points before transit, front of the star
    w_aft=np.where((xp>0) & (zp>0))[0]

    #Oblate star    
    if star_params['f_GD']>0.:

        #Frame conversion of planet coordinates from the classical frame perpendicular to the LOS, to the 'inclined star' frame
        idx_front = np_where1D(zp>0)
        xp_st_sk,yp_st_sk,_=frameconv_skyorb_to_skystar(pl_params['lambda_rad'],xp[idx_front],yp[idx_front],None)              

        #Number of planet limb points within the projected stellar photosphere
        nlimb = 501
        nlimb_in_ph = pl_limb_in_oblate_star(nlimb,RpRs,xp_st_sk,yp_st_sk,star_params)

        #First and fourth contacts: start of ingress / end of egress
        #    nlimb_in_ph >0 for the first / last time
        w_first=np_where1D(nlimb_in_ph[w_bef]>0)[0]
        w_fth=np_where1D(nlimb_in_ph[w_aft]>0)[-1]
        
        #Second and third contacts: end of ingress / start of egress
        #    nlimb_in_ph = nlimb for the first / last time
        w_scnd=np_where1D(nlimb_in_ph[w_bef]==nlimb)[0]
        w_thrd=np_where1D(nlimb_in_ph[w_aft]==nlimb)[-1]


    #Spherical star:
    else: 
        
        #Distance star - planet centers in the plane of sky
        Dprojplanet=np.sqrt(np.power(xp,2.) + np.power(yp,2.))
    
        #First contact: start of ingress
        #    dist_p = Rs+Rp
        w_first=closest(Dprojplanet[w_bef],(1.+RpRs))
    
        #Second contact: end of ingress
        #    dist_p = Rs-Rp
        w_scnd=closest(Dprojplanet[w_bef],(1.-RpRs))
    
        #Third contact: start of egress
        #    dist_p = Rs-Rp
        w_thrd=closest(Dprojplanet[w_aft],(1.-RpRs))

    
        #Fourth contact: end of egress
        #    dist_p = Rs+Rp
        w_fth=closest(Dprojplanet[w_aft],(1.+RpRs))
    
    #Contacts
    if w_bef[w_first]==0:stop('Start phase is too short for contact determination: increase "plot_dic["stend_ph"]"')
    if w_aft[w_fth]==n_pts_contacts:stop('End phase is too short for contact determination: increase "plot_dic["stend_ph"]"')
    contact_phases[0]=ph_contacts[w_bef][w_first]      
    contact_phases[1]=ph_contacts[w_bef][w_scnd]     
    contact_phases[2]=ph_contacts[w_aft][w_thrd]     
    contact_phases[3]=ph_contacts[w_aft][w_fth]    
        
    return contact_phases   



def eclipse_status(Dprojp_all,RpRs,lambda_rad,star_params,xp_sk_all,yp_sk_all,zp_sk_all):
    r"""**Exposure eclipse status.**

    Identify planet transit and occulation status at a given orbital position.
    
     - Post-occultation/pre-transit: ecl = -1
     - Ingress of transit: ecl = +2
     - Full transit: ecl = +3
     - Egress of transit: ecl = +2
     - Post-transit/pre-occultation: ecl = +1
     - Ingress of occultation: ecl = -2
     - Full occultation: ecl = -3
     - Egress of occultation: ecl = -2

    To be conservative, an exposure is considered out of the transit/occultation, or fully in-transit/occultation if it is such during the entire exposure.
    Input planet position either correspond to the center of the exposure, or to its (start,central,end) values.
     
    Args:
        TBD
    
    Returns:
        None
    
    """ 
    dim_input = np.shape(xp_sk_all)
    if len(dim_input)==1:exact_pos = True
    else:exact_pos = False
    
    #Oblate star
    if star_params['f_GD']>0.:
        nlimb = 501
        if exact_pos:
            nexp = len(xp_sk_all)
            x_st_sky_all,y_st_sky_all,_=frameconv_skyorb_to_skystar(lambda_rad,xp_sk_all,yp_sk_all,None) 
            nlimb_in_ph = np.array([ pl_limb_in_oblate_star(nlimb,RpRs,[x_st_sky],[y_st_sky],star_params)[0] for x_st_sky,y_st_sky in zip(x_st_sky_all,y_st_sky_all) ])   
        else:
            nexp = len(xp_sk_all[0])

            #Frame conversion of planet coordinates from the classical frame perpendicular to the LOS, to the 'inclined star' frame
            start_x_st_sky_all,start_y_st_sky_all,_=frameconv_skyorb_to_skystar(lambda_rad,xp_sk_all[0],yp_sk_all[0],None)           
            end_x_st_sky_all,end_y_st_sky_all,_=frameconv_skyorb_to_skystar(lambda_rad,xp_sk_all[2],yp_sk_all[2],None) 
    
            #Number of planet limb points within the projected stellar photosphere
            nlimb_in_ph = np.zeros(nexp,dtype=float)
            for isub,(start_x_st_sky,start_y_st_sky,end_x_st_sky,end_y_st_sky) in enumerate(zip(start_x_st_sky_all,start_y_st_sky_all,end_x_st_sky_all,end_y_st_sky_all)):
                start_nlimb_in_ph = pl_limb_in_oblate_star(nlimb,RpRs,[start_x_st_sky],[start_y_st_sky],star_params)[0]
                end_nlimb_in_ph = pl_limb_in_oblate_star(nlimb,RpRs,[end_x_st_sky],[end_y_st_sky],star_params)[0]
                nlimb_in_ph[isub] = start_nlimb_in_ph+end_nlimb_in_ph
    
        #Planetary disk is outside the projected stellar photosphere
        cond_out = (nlimb_in_ph==0)

        #Planet is entirely in front of the disk  
        cond_in = (nlimb_in_ph==2*nlimb)
            
    #Spherical star    
    else:
        if exact_pos:
            nexp = len(Dprojp_all)
            cond_out = (Dprojp_all > (1.+RpRs)) 
            cond_in =  (Dprojp_all <= (1.-RpRs))  
        else:
            nexp = len(Dprojp_all[0])
            cond_out = (Dprojp_all[0] > (1.+RpRs)) & (Dprojp_all[2] > (1.+RpRs))
            cond_in = (Dprojp_all[0] <= (1.-RpRs)) & (Dprojp_all[2] <= (1.-RpRs))
       
    #Planet is in front or behind the sky-projected plane
    if exact_pos:cond_tr = (zp_sk_all>0.)
    else:cond_tr = (0.5*(zp_sk_all[0]+zp_sk_all[2])>0.)
    cond_occ = ~cond_tr
    
    #Default (ingress/egress of transit and occultation)
    eclipse = np.repeat(2.,nexp)      
    eclipse[cond_occ]*=-1.  
    
    #Planet is entirely in front of the disk (full transit) 
    eclipse[cond_in & cond_tr] = 3.

    #Planet is entirely behind the disk (full occultation) 
    eclipse[cond_in & cond_occ] = -3.    

    #Planetary disk is outside the projected stellar photosphere
    eclipse[cond_out] = 1.

    #Post-occultation and pre-transit 
    if exact_pos:eclipse[xp_sk_all<0.]*=-1.
    else:eclipse[0.5*(xp_sk_all[0]+xp_sk_all[2])<0.]*=-1.
                                    
    return eclipse


def pl_limb_in_oblate_star(nlimb,RpRs,xp_st_sk,yp_st_sk,star_params):
    r"""**Transit contacts.**

    Calculates the number of discretized planet limb points within the projected oblate stellar photosphere, for a given orbital position.

    Args:
        TBD
    
    Returns:
        None
    
    """ 
    xlimb = RpRs*np.cos(2*np.pi*np.arange(nlimb)/(nlimb-1.))
    ylimb = RpRs*np.sin(2*np.pi*np.arange(nlimb)/(nlimb-1.))
    nlimb_in_ph = np.zeros(len(xp_st_sk),dtype=int)
    for iloc,(xp_st_sk_loc,yp_st_sk_loc) in enumerate(zip(xp_st_sk,yp_st_sk)):
        x_st_sk_limb = xp_st_sk_loc+xlimb
        y_st_sk_limb = yp_st_sk_loc+ylimb
        cond_in_stphot=calc_zLOS_oblate(x_st_sk_limb,y_st_sk_limb,star_params['istar_rad'],star_params['RpoleReq'])[2] 
        nlimb_in_ph[iloc] = np.sum( cond_in_stphot )    
    return nlimb_in_ph















##################################################################################################    
#%% Frame change
################################################################################################## 

def frameconv_skyorb_to_skystar(lambd,xin,yin,zin):
    r"""**Frame: sky-projected, orbital to stellar.**

    Converts coordinates from the classical sky-projected orbital frame, to the sky-projected (inclined) stellar frame.
    Frame is rotated around the :math:`Z_\mathrm{LOS}` axis (the LOS, which remains unchanged), by angle :math:`\lambda` counted :math:`>0` counterclockwise
    from the :math:`X_\mathrm{sky,star}` to the :math:`X_\mathrm{sky}` axis, or from the :math:`Y_\mathrm{sky,star}` to the :math:`Y_\mathrm{sky}` axis, in the :math:`XY_\mathrm{sky}` plane.

    .. math:: 
       X_\mathrm{sky,star} &= \mathrm{sky-projected \, stellar \, equator}  \\
       Y_\mathrm{sky,star} &= \mathrm{sky-projected \, stellar \, spin} \\
       Z_\mathrm{sky,star} &= \mathrm{LOS}

    Args:
        TBD
    
    Returns:
        None
    
    """ 
    xout = xin*np.cos(lambd) - yin*np.sin(lambd)
    yout = xin*np.sin(lambd) + yin*np.cos(lambd)
    zout=deepcopy(zin)
    return xout,yout,zout

def frameconv_skystar_to_skyorb(lambd,xin,yin,zin):
    r"""**Frame: sky-projected, stellar to orbital.**

    Converts coordinates from the sky-projected (inclined) stellar frame to the classical sky-projected orbital frame.

    Args:
        TBD
    
    Returns:
        None
    
    """ 
    xout = xin*np.cos(lambd)+yin*np.sin(lambd) 
    yout = -xin*np.sin(lambd)+yin*np.cos(lambd) 
    zout=deepcopy(zin)  
    return xout,yout,zout



def frameconv_skystar_to_star(xin,yin,zin,istar):
    r"""**Frame: sky-projected stellar to stellar.**

    Converts coordinates from the sky-projected (inclined) stellar frame to the star frame. 
    Frame is rotated around the :math:`X_\mathrm{sky,star}` axis (which remains unchanged), by angle :math:`i_{\star}` counted :math:`>0` counted counterclockwise
    from the :math:`Z_\mathrm{sky,star}` to the :math:`Y_\star` axis in the :math:`YZ_\mathrm{sky,star}` plane.

    .. math:: 
       X_\mathrm{star} &= \mathrm{star \,equator} \\
       Y_\mathrm{star} &= \mathrm{star \, spin} \\
       Z_\mathrm{star} &= \mathrm{complements \, frame}  

    Args:
        TBD
    
    Returns:
        None
    
    """ 
    xout=deepcopy(xin)
    yout= np.sin(istar)*yin + np.cos(istar)*zin
    zout=-np.cos(istar)*yin + np.sin(istar)*zin   
    return xout,yout,zout

def frameconv_star_to_skystar(xin,yin,zin,istar):
    r"""**Frame: stellar to  sky-projected stellar.**

    Converts coordinates from the star frame to the sky-projected (inclined) stellar frame.

    Args:
        TBD
    
    Returns:
        None
    
    """ 
    xout=deepcopy(xin)
    yout= np.sin(istar)*yin - np.cos(istar)*zin
    zout= np.cos(istar)*yin + np.sin(istar)*zin   
    return xout,yout,zout








#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os as os_system
import argparse
import json
import logging
#The following relative imports are necessary to create an executable command
from ..ANTARESS_process.ANTARESS_main import ANTARESS_main,ANTARESS_settings_overwrite
from ..ANTARESS_launch.ANTARESS_gridrun import ANTARESS_gridrun
from ..ANTARESS_general.utils import import_module
 

def ANTARESS_launcher(custom_systems = '',custom_settings = '',custom_plot_settings = '',working_path='',nbook_dic = {} , exec_comm = True):
    r"""**ANTARESS launch routine.**
    
    Runs ANTARESS with default or manual settings.  
    
    Args:
        custom_systems (str): name of custom systems file (default "": ANTARESS_systems.py file is used)
        custom_settings (str): name of custom settings file (default "": ANTARESS_settings.py file is used)
        custom_plot_settings (str): name of custom plot settings file (default "": ANTARESS_plot_settings.py file is used)
        working_path (str): path to the working directory, in which the workflow outputs will be stored (default "": current directory is used)
                            if custom files are used, they should be placed in the working directory 
    
    Returns:
        None
    
    """ 
    
    #Suppress log messages from the fontTools package
    fontTools_logger = logging.getLogger('fontTools')
    fontTools_logger.addHandler(logging.NullHandler())
    
    #Read executable arguments
    #    - will be used when ANTARESS is called as an executable through terminal
    if exec_comm:
        parser=argparse.ArgumentParser(prog = "antaress",description='Launch ANTARESS workflow')
        parser.add_argument("--custom_systems",      type=str, default='',help = 'Name of custom systems file (default "": default file ANTARESS_systems.py is used)')
        parser.add_argument("--custom_settings",     type=str, default='',help = 'Name of custom settings file (default "": default file ANTARESS_settings.py is used)')
        parser.add_argument("--custom_plot_settings",type=str, default='',help = 'Name of custom plot settings file (default "": default file ANTARESS_plot_settings.py is used)')
        parser.add_argument("--working_path", type=str, default='' ,help = 'Path to user settings files (default "./": user files are retrieved from current directory)')
        parser.add_argument("-d", "--nbook_dic", type=json.loads, default={})
        input_args=parser.parse_args()
        custom_systems = input_args.custom_systems
        custom_settings = input_args.custom_settings
        custom_plot_settings = input_args.custom_plot_settings
        working_path = input_args.working_path
        nbook_dic = input_args.nbook_dic 

    #Initializes main dictionaries
    gen_dic={}
    plot_dic={}
    corr_spot_dic={}
    data_dic={
        'DI':{'fit_prof':{},'mask':{}},
        'Res':{},
        'PCA':{},
        'Intr':{'fit_prof':{},'mask':{}},
        'Atm':{'fit_prof':{},'mask':{}}}
    mock_dic={}
    theo_dic={}
    detrend_prof_dic={}
    glob_fit_dic={
        'DIProp':{},
        'IntrProp':{},
        'ResProf':{},
        'IntrProf':{},
        'AtmProp':{},
        'AtmProf':{},
        }  
    
    #Retrieve default settings
    from ..ANTARESS_launch.ANTARESS_settings import ANTARESS_settings
    ANTARESS_settings(gen_dic,plot_dic,corr_spot_dic,data_dic,mock_dic,theo_dic,glob_fit_dic,detrend_prof_dic)
    
    #Overwrite with user settings
    if custom_settings!='':import_module(working_path+custom_settings).ANTARESS_settings(gen_dic,plot_dic,corr_spot_dic,data_dic,mock_dic,theo_dic,glob_fit_dic,detrend_prof_dic)

    #Overwrite with notebook settings
    if ('settings' in nbook_dic) and (len(nbook_dic['settings'])>0):
        ANTARESS_settings_overwrite(gen_dic,plot_dic,corr_spot_dic,data_dic,mock_dic,theo_dic,glob_fit_dic,detrend_prof_dic,nbook_dic)

    #----------------------------------------------------------------------------------------------------    

    #Working directory
    if working_path=='':gen_dic['save_dir'] = os_system.getcwd()+'/'
    else:gen_dic['save_dir']= working_path
    
    #Code directory     
    code_dir = os_system.path.dirname(__file__).split('ANTARESS_launch')[0]

    #Retrieve default or user-defined system properties
    if custom_systems!='':systems_file = import_module(working_path+custom_systems)
    else:systems_file = import_module(code_dir+'ANTARESS_launch/ANTARESS_systems.py')
    all_system_params = systems_file.get_system_params()

    #Overwrite with notebook settings
    if ('system' in nbook_dic) and (len(nbook_dic['system'])>0):
        all_system_params.update(nbook_dic['system'])
        
    #Retrieve chosen system
    system_params = all_system_params[gen_dic['star_name']]

    print('****************************************')
    print('Launching ANTARESS')
    print('****************************************')
    print('')

    #Moving to code directory
    os_system.chdir(code_dir)
  
    #Run over nominal settings properties
    #    - notebook settings have already been used to overwrite congiguration settings, and are only passed on to overwrite the plot settings if relevant
    if not gen_dic['grid_run']:
        ANTARESS_main(data_dic,mock_dic,gen_dic,theo_dic,plot_dic,glob_fit_dic,detrend_prof_dic, corr_spot_dic,system_params,nbook_dic,custom_plot_settings)
    
    #Run over a grid of properties
    #    - will overwrite default and notebook configuration settings
    else:
        ANTARESS_gridrun(data_dic,mock_dic,gen_dic,theo_dic,plot_dic,glob_fit_dic,detrend_prof_dic, corr_spot_dic,system_params,nbook_dic,custom_plot_settings)

    print('End of workflow')
    return None










#!/usr/bin/env python3
# -*- coding: utf-8 -*-




'''
Wrap-up function to process atmospheric profiles
'''
def joined_Atm_ana(gen_dic):

    #Fitting intrinsic stellar CCFs with joined model
    if gen_dic['fit_AtmProf']:
        fit_CCFatm_all()    
        
    #Fitting stellar surface properties with a linked model
    if gen_dic['fit_AtmProp']:
        fit_CCFatm_prop()    
    

    return None




'''
Routine to fit atmospheric CCFs with a joined model
    - we use simple analytical profiles to describe the CCFs
      their shape and position are linked using polynomial laws as a function of phase
    - contrast and FWHM correspond to the atmospheric line before instrumental convolution, so that data from all instruments and visits can be fitted together
    - polynomial coefficients can nonetheless depend on the visit and their associated instrument, to account for possible variations in the line shape between visits
'''
def fit_CCFatm_all():
    print('   > Fitting atmospheric CCFs with linked model')


    return None







'''
Routine to fit atmospheric CCF properties with a common model
'''
def fit_CCFatm_prop():
    print('   > Fitting atmospheric properties with linked model')


    return None














    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy as np
from copy import deepcopy
from ..ANTARESS_general.utils import np_where1D,dataload_npz,check_data
from ..ANTARESS_corrections.ANTARESS_tellurics import corr_tell
from ..ANTARESS_corrections.ANTARESS_flux_balance import def_Mstar,corr_Fbal,corr_Ftemp
from ..ANTARESS_corrections.ANTARESS_peaks import corr_cosm,MAIN_permpeak
from ..ANTARESS_corrections.ANTARESS_interferences import MAIN_corr_wig,corr_fring

def red_sp_data_instru(inst,data_dic,plot_dic,gen_dic,data_prop,coord_dic,system_param):
    r"""**Correction sequence for spectral data.**
    
    Calls the successive modules correcting spectra for environmental/instrumental effects
    All datasets from a given instrument are processed together, as some corrections benefit from combining several visits, and from using the full range of the input data. 
    After the correction sequence, spectra can be trimmed to a chosen set of spectral orders and/or spectral bands if the subsequent analysis focuses on a specific region. 
    
    Args:
        TBD
    
    Returns:
        None
    
    """
    #Information
    #    - plots are deactivated if corrections are not required 
    #    - every time data is corrected in a module the path to the 'processed' data is updated, so that later modules will use the most corrected data
    data_inst=data_dic[inst]

    #Save path to uncorrected data
    for vis in data_inst['visit_list']:data_inst[vis]['uncorr_exp_data_paths']=deepcopy(data_inst[vis]['proc_DI_data_paths'])

    #Correcting for tellurics
    if (gen_dic['corr_tell']):corr_tell(gen_dic,data_inst,inst,data_dic,data_prop,coord_dic,plot_dic)  

    #Definition of a global master for the star
    if (gen_dic['glob_mast']):def_Mstar(gen_dic,data_inst,inst,data_prop,plot_dic,data_dic,coord_dic)

    #Correcting for flux balance
    if gen_dic['corr_Fbal'] or gen_dic['corr_FbalOrd']:corr_Fbal(inst,gen_dic,data_inst,plot_dic,data_prop,data_dic)  

    #Correcting for temporal variations
    if (gen_dic['corr_Ftemp']):corr_Ftemp(inst,gen_dic,data_inst,plot_dic,data_prop,coord_dic,data_dic) 

    #Correcting for cosmics
    if (gen_dic['corr_cosm']):corr_cosm(inst,gen_dic,data_inst,plot_dic,data_dic,coord_dic)

    #Masking of persistent features
    if (gen_dic['mask_permpeak']):MAIN_permpeak(inst,gen_dic,data_inst,plot_dic,data_dic,data_prop)

    #Correcting for ESPRESSO wiggles
    if (gen_dic['corr_wig']):MAIN_corr_wig(inst,gen_dic,data_dic,coord_dic,data_prop,plot_dic,system_param)

    #Correcting for fringing
    if (gen_dic['corr_fring']):corr_fring(inst,gen_dic,data_inst,plot_dic,data_dic) 

    #Reducing spectra to the selected spectral ranges
    if (gen_dic['trim_spec']):lim_sp_range(inst,data_dic,gen_dic,data_prop)

    #Save path to corrected data
    for vis in data_inst['visit_list']:data_inst[vis]['corr_exp_data_paths']=deepcopy(data_inst[vis]['proc_DI_data_paths'])

    return None





def lim_sp_range(inst,data_dic,gen_dic,data_prop):
    r"""**Data trimming.**
    
    Limits input spectra to specific spectral ranges and order

     - orders left empty in all visits are removed
    
    Args:
        TBD
    
    Returns:
        None
    
    """
    print('   > Trimming spectra')    
 
    #Calculating data
    data_inst = data_dic[inst]
    if (gen_dic['calc_trim_spec']):
        print('         Calculating data')    
    
        #Upload latest processed data
        #    - data must be put in global tables to perform a global reduction of ranges and orders
        data_dic_exp={}
        data_dic_com={}
        data_com_inst = dataload_npz(gen_dic['save_data_dir']+'Processed_data/'+inst+'_com') 
        for vis in data_inst['visit_list']:    
            data_vis=data_inst[vis]
            data_dic_com[vis] = dataload_npz(data_vis['proc_com_data_paths'])
            data_dic_exp[vis] = {}            
            for key in ['cen_bins','flux','mean_gdet']:data_dic_exp[vis][key]=np.zeros(data_vis['dim_all'], dtype=float)*np.nan
            data_dic_exp[vis]['cond_def']=np.zeros(data_vis['dim_all'], dtype=bool)
            data_dic_exp[vis]['cov'] = np.zeros(data_vis['dim_sp'], dtype=object)
            data_dic_exp[vis]['edge_bins']=np.zeros(data_vis['dim_sp']+[data_vis['nspec']+1], dtype=float)*np.nan
            if data_vis['tell_sp']:data_dic_exp[vis]['tell']=np.zeros(data_vis['dim_all'], dtype=float)
            for iexp in range(data_vis['n_in_visit']):               
                data_exp = dataload_npz(data_vis['proc_DI_data_paths']+str(iexp))                 
                for key in [key_sub for key_sub in data_dic_exp[vis] if key_sub not in ['mean_gdet','tell']]:data_dic_exp[vis][key][iexp] = data_exp[key]
                if data_vis['tell_sp']:data_dic_exp[vis]['tell'][iexp] = dataload_npz(data_vis['tell_DI_data_paths'][iexp])['tell']             
                data_dic_exp[vis]['mean_gdet'][iexp] = dataload_npz(data_vis['mean_gdet_DI_data_paths'][iexp])['mean_gdet']

        #------------------------------------------

        #Processing each visit
        if len(gen_dic['trim_range'])>0:
            if (data_inst['type']=='spec2D'):cond_ord_kept = np.repeat(False,data_inst['nord'])
            for vis in data_inst['visit_list']:    
                data_vis=data_inst[vis]
    
                #Set bins outside selected ranges to nan in flux tables
                for iexp in range(data_vis['n_in_visit']):
                    low_bins_exp = data_dic_exp[vis]['edge_bins'][iexp,:,0:-1]
                    high_bins_exp = data_dic_exp[vis]['edge_bins'][iexp,:,1::]
                    cond_keep_sp = np.zeros(data_vis['dim_exp'],dtype=bool)   
                    for trim_range_loc in gen_dic['trim_range']:
                        cond_keep_sp |= (low_bins_exp>=trim_range_loc[0]) & (high_bins_exp<=trim_range_loc[1])
                    data_dic_exp[vis]['flux'][iexp,~cond_keep_sp] = np.nan         
                    data_dic_exp[vis]['cond_def'][iexp]&=cond_keep_sp                        
               
                #Limiting all exposures to the smallest common defined range
                #    - the range is delimited by the most extreme bins with flux value defined in at least one order and one exposure
                #    - spectral tables remain continuous
                cond_def_vis = (np.sum(data_dic_exp[vis]['cond_def'],axis=(0,1)) > 0.)         
                idx_def_raw_all = np_where1D(cond_def_vis)
                if len(idx_def_raw_all)<data_vis['nspec']:
                    idx_range_kept = np.arange(idx_def_raw_all[0],idx_def_raw_all[-1]+1)  
                    for key in ['cen_bins','flux','cond_def','mean_gdet']:
                        data_dic_exp[vis][key] = np.take(data_dic_exp[vis][key],idx_range_kept,axis=-1)
                    if data_vis['tell_sp']:data_dic_exp[vis]['tell'] = np.take(data_dic_exp[vis]['tell'],idx_range_kept,axis=-1)
                    data_dic_com[vis]['cen_bins'] = np.take(data_dic_com[vis]['cen_bins'],idx_range_kept,axis=-1)
                    data_dic_exp[vis]['edge_bins'] = np.append(data_dic_exp[vis]['edge_bins'][:,:,idx_range_kept],data_dic_exp[vis]['edge_bins'][:,:,idx_range_kept[-1]+1][:,:,None],axis=2)                 
                    data_dic_com[vis]['edge_bins'] = np.append(data_dic_com[vis]['edge_bins'][:,idx_range_kept],data_dic_com[vis]['edge_bins'][:,idx_range_kept[-1]+1][:,None],axis=1)                 
                    for iexp in range(data_vis['n_in_visit']):          
                        for iord in range(data_inst['nord']):data_dic_exp[vis]['cov'][iexp,iord] = data_dic_exp[vis]['cov'][iexp,iord][:,idx_range_kept]
                    data_vis['nspec'] = len(idx_range_kept)
                    data_vis['dim_all'][2] = data_vis['nspec']
                    data_vis['dim_exp'][1] = data_vis['nspec']            
            
                #Limiting all exposures to the smallest common defined orders
                #    - some of the orders might be entirely empty after spectral ranges exclusion
                #    - an order is kept if it contains at least one defined bin in at least one exposure
                if (data_inst['type']=='spec2D'):cond_ord_kept |= (np.sum(data_dic_exp[vis]['cond_def'],axis=(0,2)) > 0.)
    
            #Limiting inst common table to smallest defined range
            cond_keep_sp = np.zeros(data_inst['dim_exp'],dtype=bool)  
            for trim_range_loc in gen_dic['trim_range']:
                cond_keep_sp |= (data_com_inst['edge_bins'][:,0:-1]>=trim_range_loc[0]) & (data_com_inst['edge_bins'][:,1::]<=trim_range_loc[1]) 
            idx_def_raw_inst = np_where1D(np.sum(cond_keep_sp,axis=(0)) > 0.)
            if len(idx_def_raw_inst)<data_inst['nspec']:
                idx_range_kept = np.arange(idx_def_raw_inst[0],idx_def_raw_inst[-1]+1) 
                data_com_inst['cen_bins'] = np.take(data_com_inst['cen_bins'],idx_range_kept,axis=-1)
                data_com_inst['edge_bins'] = np.append(data_com_inst['edge_bins'][:,idx_range_kept],data_com_inst['edge_bins'][:,idx_range_kept[-1]+1][:,None],axis=1)                 
                data_inst['nspec'] = len(idx_range_kept)               
                data_inst['dim_exp'][1] = data_inst['nspec']   
        elif (data_inst['type']=='spec2D'):
            cond_ord_kept = np.repeat(True,data_inst['nord'])    #condition relative to current tables
        
        #Reduce 2D tables to common defined, and selected, orders over all visits
        #    - an order has been kept if it is defined in at least one visit
        if (data_inst['type']=='spec2D'):
            idx_ord_kept = np_where1D(cond_ord_kept)             #indexes relative to current tables
            if (inst in gen_dic['trim_orders']) and (len(gen_dic['trim_orders'][inst])>0):
                idx_ord_kept = np.intersect1d(idx_ord_kept,gen_dic['trim_orders'][inst])
            if len(idx_ord_kept)<data_inst['nord']:
                cond_orders4ccf = np.zeros(data_inst['nord'],dtype=bool)
                cond_orders4ccf[gen_dic[inst]['orders4ccf']] = True
                data_inst['nord'] = len(idx_ord_kept)  
                data_inst['nord_spec']=deepcopy(data_inst['nord'])     
                gen_dic[inst]['orders4ccf'] = np_where1D(cond_orders4ccf[idx_ord_kept]) #index of orders relative to the new reduced tables    
                data_inst['idx_ord_ref'] = np.array(data_inst['idx_ord_ref'])[idx_ord_kept]
                for key in ['cen_bins','edge_bins']:
                    data_com_inst[key] = np.take(data_com_inst[key],idx_ord_kept,axis=0)    
                data_inst['dim_exp'][0] = data_inst['nord']  
                for vis in data_inst['visit_list']: 
                    data_vis=data_inst[vis] 
                    for key in data_dic_exp[vis]:
                        data_dic_exp[vis][key] = np.take(data_dic_exp[vis][key],idx_ord_kept,axis=1) 
                    for key in ['cen_bins','edge_bins']:
                        data_dic_com[vis][key] = np.take(data_dic_com[vis][key],idx_ord_kept,axis=0)
                    data_vis['dim_all'][1] = data_inst['nord']
                    data_vis['dim_exp'][0] = data_inst['nord']
                    data_vis['dim_sp'][1] = data_inst['nord']     

        #Saving modified data and updating paths
        data_inst['proc_com_data_path'] = gen_dic['save_data_dir']+'Corr_data/Trim/'+inst+'_com' 
        np.savez_compressed(data_inst['proc_com_data_path'],data = data_com_inst,allow_pickle=True)    
        np.savez(gen_dic['save_data_dir']+'Corr_data/Trim/DimTrimmed_'+inst,nord = data_inst['nord'],nspec = data_inst['nspec'],dim_exp = data_inst['dim_exp'],orders4ccf = gen_dic[inst]['orders4ccf'],idx_ord_ref=data_inst['idx_ord_ref'],nord_spec=data_inst['nord_spec'])  
        for vis in data_inst['visit_list']:
            data_vis=data_inst[vis]
            data_vis['proc_DI_data_paths'] = gen_dic['save_data_dir']+'Corr_data/Trim/'+inst+'_'+vis+'_'
            if data_vis['tell_sp']:data_vis['tell_DI_data_paths'] = {}
            data_vis['mean_gdet_DI_data_paths'] = {}
            for iexp in range(data_vis['n_in_visit']): 
                if data_vis['tell_sp']:
                    data_vis['tell_DI_data_paths'][iexp] = gen_dic['save_data_dir']+'Corr_data/Trim/'+inst+'_'+vis+'_tell_'+str(iexp)
                    np.savez_compressed(data_vis['tell_DI_data_paths'][iexp],data={'tell':data_dic_exp[vis]['tell'][iexp]},allow_pickle=True)
                data_vis['mean_gdet_DI_data_paths'][iexp] = gen_dic['save_data_dir']+'Corr_data/Trim/'+inst+'_'+vis+'_mean_gdet_'+str(iexp)
                np.savez_compressed(data_vis['mean_gdet_DI_data_paths'][iexp],data={'mean_gdet':data_dic_exp[vis]['mean_gdet'][iexp]},allow_pickle=True)   
                data_sav_exp = {key:data_dic_exp[vis][key][iexp] for key in data_dic_exp[vis] if key not in ['tell','mean_gdet']}                
                np.savez_compressed(data_vis['proc_DI_data_paths']+str(iexp),data=data_sav_exp,allow_pickle=True)
            data_vis['proc_com_data_paths'] = gen_dic['save_data_dir']+'Corr_data/Trim/'+inst+'_'+vis+'_com'          
            np.savez_compressed(data_vis['proc_com_data_paths'],data = data_dic_com[vis],allow_pickle=True)    
            np.savez(gen_dic['save_data_dir']+'Corr_data/Trim/DimTrimmed_'+inst+'_'+vis,
                     nspec = data_vis['nspec'],dim_exp = data_vis['dim_exp'],dim_sp = data_vis['dim_sp'],dim_all = data_vis['dim_all'])     
    
    #Updating path to processed data and checking it has been calculated
    else:
        data_inst['proc_com_data_path'] = gen_dic['save_data_dir']+'Corr_data/Trim/'+inst+'_com'
        data_load = np.load(gen_dic['save_data_dir']+'Corr_data/Trim/DimTrimmed_'+inst+'.npz')
        data_inst['dim_exp'] = list(data_load['dim_exp'])
        for key in ['nord','nspec','idx_ord_ref','nord_spec']:data_inst[key]=data_load[key]
        gen_dic[inst]['orders4ccf'] = data_load['orders4ccf']                 
        for vis in data_inst['visit_list']: 
            data_vis=data_inst[vis]
            data_vis['proc_com_data_paths'] = gen_dic['save_data_dir']+'Corr_data/Trim/'+inst+'_'+vis+'_com'
            check_data({'path':data_vis['proc_com_data_paths']},vis=vis)
            data_vis['proc_DI_data_paths'] = gen_dic['save_data_dir']+'Corr_data/Trim/'+inst+'_'+vis+'_'
            if data_vis['tell_sp']:data_vis['tell_DI_data_paths']={}
            data_vis['mean_gdet_DI_data_paths'] = {}
            for iexp in range(data_vis['n_in_visit']): 
                if data_vis['tell_sp']:
                    data_vis['tell_DI_data_paths'][iexp] = gen_dic['save_data_dir']+'Corr_data/Trim/'+inst+'_'+vis+'_tell_'+str(iexp)
                data_vis['mean_gdet_DI_data_paths'][iexp] = gen_dic['save_data_dir']+'Corr_data/Trim/'+inst+'_'+vis+'_mean_gdet_'+str(iexp)
            data_load = np.load(gen_dic['save_data_dir']+'Corr_data/Trim/DimTrimmed_'+inst+'_'+vis+'.npz')
            for key in ['dim_all','dim_exp','dim_sp']:data_inst[vis][key] = list(data_load[key])
            data_inst[vis]['nspec']=data_load['nspec']  

    return None













#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy as np
from ..ANTARESS_process.ANTARESS_main import ANTARESS_main,ANTARESS_settings_overwrite

def ANTARESS_gridrun(data_dic,mock_dic,gen_dic,theo_dic,plot_dic,glob_fit_dic,detrend_prof_dic, corr_spot_dic,system_params,nbook_dic,input_path,custom_plot_settings):
    r"""**ANTARESS grid run launch.**
    
    Runs ANTARESS over a grid of values for a given set of fields in the configuration file.  
    
    Args:
        TBD
    
    Returns:
        None
    
    """ 
    grid_prop = {}

    #Grid over individual spectral orders
    iord_run = np.delete(np.arange(85),[59,68,74,75,76,77,78,82])    #empty CCF orders (tellurics)
    for iord in iord_run: grid_prop['orders4ccf'] += [2*iord,2*iord+1]   
   
    # inst = 'HARPN'
    # iord_run = np.delete(np.arange(69),[54,63,68])     #Old masks 

    # inst = 'HARPN'
    # iord_run = np.delete(np.arange(69),[53,54,60,63,64])     #New masks 

    # for iord in iord_run: grid_prop['orders4ccf'] += [iord] 

    #Instrument
    inst = 'ESPRESSO'


    
    
    #------------------------------------------------------------------------------------------------

    #Grid preparation
    par_names = []
    grid_par = []
    for prop in grid_prop:
        par_names+=[prop]
        grid_par+=[list(grid_prop[prop])]
    grid_par = np.array(grid_par).T

    #Looping on grid 
    gridstep_dic = {
        'orders4ccf':{},    
    }

    #Run the code over grid
    for par_set in grid_par:    
        for par_name,par_val in zip(par_names,par_set):
            root_name = par_name

            #Spectral orders
            if root_name == 'orders4ccf':
                gridstep_dic['orders4ccf'][inst] = par_val 
    
            #----------------------------------------------------------------------------------------------------------------------------------   
            
            #Overwrite current settings        
            ANTARESS_settings_overwrite(gen_dic,plot_dic,corr_spot_dic,data_dic,mock_dic,theo_dic,glob_fit_dic,detrend_prof_dic,gridstep_dic)
            
            #Run code for current parameter set        
            ANTARESS_main(data_dic,mock_dic,gen_dic,theo_dic,plot_dic,glob_fit_dic,detrend_prof_dic, corr_spot_dic,system_params,nbook_dic,custom_plot_settings)


    print('End of grid')
    return None
